"""
Estimation and modelling of spatial variograms
==============================================

Digital elevation models have errors that are often `correlated in space <https://en.wikipedia.org/wiki/Spatial_analysis#Spatial_auto-correlation>`_.
While many DEM studies used solely short-range `variogram <https://en.wikipedia.org/wiki/Variogram>`_ to
estimate the correlation of elevation measurement errors (e.g., `Howat et al. (2008) <https://doi.org/10.1029/2008GL034496>`_
, `Wang and Kääb (2015) <https://doi.org/10.3390/rs70810117>`_), recent studies show that variograms of multiple ranges
provide larger, more reliable estimates of spatial correlation for DEMs.

Here, we show an example in which we estimate the spatial correlation for a DEM difference at Longyearbyen, and its
impact on the standard error with averaging area. We first estimate an empirical variogram with
:func:`xdem.spatialstats.sample_empirical_variogram` based on routines of `scikit-gstat
<https://mmaelicke.github.io/scikit-gstat/index.html>`_. We then fit the empirical variogram with a sum of variogram
models using :func:`xdem.spatialstats.fit_sum_model_variogram`. Finally, we perform spatial propagation for a range of
averaging area using :func:`xdem.spatialstats.number_effective_samples`, and empirically validate the improved
robustness of our results using :func:`xdem.spatialstats.patches_method`, an intensive Monte-Carlo sampling approach.

**Reference:** `Hugonnet et al. (2022) <https://doi.org/10.1109/jstars.2022.3188922>`_, Figure 5 and Equations 13–16.
"""
import geoutils as gu

# sphinx_gallery_thumbnail_number = 6
import matplotlib.pyplot as plt
import numpy as np

import xdem
from xdem.spatialstats import nmad

# %%
# We load example files.

dh = xdem.DEM(xdem.examples.get_path("longyearbyen_ddem"))
glacier_outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
mask_glacier = glacier_outlines.create_mask(dh)

# %%
# We exclude values on glacier terrain in order to isolate stable terrain, our proxy for elevation errors.
dh.set_mask(mask_glacier)

# %%
# We estimate the average per-pixel elevation error on stable terrain, using both the standard deviation
# and normalized median absolute deviation. For this example, we do not account for elevation heteroscedasticity.
print(f"STD: {np.nanstd(dh.data):.2f} meters.")
print(f"NMAD: {xdem.spatialstats.nmad(dh.data):.2f} meters.")

# %%
# The two measures of dispersion are quite similar showing that, on average, there is a small influence of outliers on the
# elevation differences. The per-pixel precision is about :math:`\pm` 2.5 meters.
# **Does this mean that every pixel has an independent measurement error of** :math:`\pm` **2.5 meters?**
# Let's plot the elevation differences to visually check the quality of the data.
plt.figure(figsize=(8, 5))
dh.show(ax=plt.gca(), cmap="RdYlBu", vmin=-4, vmax=4, cbar_title="Elevation differences (m)")

# %%
# We clearly see that the residual elevation differences on stable terrain are not random. The positive and negative
# differences (blue and red, respectively) appear correlated over large distances. These correlated errors are what
# we want to estimate and model.

# %%
# Additionally, we notice that the elevation differences are still polluted by unrealistically large elevation
# differences near glaciers, probably because the glacier inventory is more recent than the data, hence with too small outlines.
# To remedy this, we filter large elevation differences outside 4 NMAD.
dh.set_mask(np.abs(dh.data) > 4 * xdem.spatialstats.nmad(dh.data))

# %%
# We plot the elevation differences after filtering to check that we successively removed glacier signals.
plt.figure(figsize=(8, 5))
dh.show(ax=plt.gca(), cmap="RdYlBu", vmin=-4, vmax=4, cbar_title="Elevation differences (m)")

# %%
# To quantify the spatial correlation of the data, we sample an empirical variogram.
# The empirical variogram describes the variance between the elevation differences of pairs of pixels depending on their
# distance. This distance between pairs of pixels if referred to as spatial lag.
#
# To perform this procedure effectively, we use improved methods that provide efficient pairwise sampling methods for
# large grid data in `scikit-gstat <https://mmaelicke.github.io/scikit-gstat/index.html>`_, which are encapsulated
# conveniently by :func:`xdem.spatialstats.sample_empirical_variogram`:
# Dowd's variogram is used for robustness in conjunction with the NMAD (see :ref:`robuststats-corr`).

df = xdem.spatialstats.sample_empirical_variogram(
    values=dh.data, gsd=dh.res[0], subsample=100, n_variograms=10, estimator="dowd", random_state=42
)

# %%
# *Note: in this example, we add a* ``random_state`` *argument to yield a reproducible random sampling of pixels within
# the grid.*

# %%
# We plot the empirical variogram:
xdem.spatialstats.plot_variogram(df)

# %%
# With this plot, it is hard to conclude anything! Properly visualizing the empirical variogram is one of the most
# important step. With grid data, we expect short-range correlations close to the resolution of the grid (~20-200
# meters), but also possibly longer range correlation due to instrument noise or alignment issues (~1-50 km).
#
# To better visualize the variogram, we can either change the axis to log-scale, but this might make it more difficult
# to later compare to variogram models. # Another solution is to split the variogram plot into subpanels, each with
# its own linear scale. Both are shown below.

# %%
# **Log scale:**
xdem.spatialstats.plot_variogram(df, xscale="log")

# %%
# **Subpanels with linear scale:**
xdem.spatialstats.plot_variogram(df, xscale_range_split=[100, 1000, 10000])

# %%
# We identify:
#   - a short-range (i.e., correlation length) correlation, likely due to effects of resolution. It has a large partial sill (correlated variance), meaning that the elevation measurement errors are strongly correlated until a range of ~100 m.
#   - a longer range correlation, with a smaller partial sill, meaning the part of the elevation measurement errors remain correlated over a longer distance.
#
# In order to show the difference between accounting only for the most noticeable, short-range correlation, or adding the
# long-range correlation, we fit this empirical variogram with two different models: a single spherical model, and
# the sum of two spherical models (two ranges). For this, we use :func:`xdem.spatialstats.fit_sum_model_variogram`, which
# is based on `scipy.optimize.curve_fit <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html>`_:
func_sum_vgm1, params_vgm1 = xdem.spatialstats.fit_sum_model_variogram(
    list_models=["Spherical"], empirical_variogram=df
)

func_sum_vgm2, params_vgm2 = xdem.spatialstats.fit_sum_model_variogram(
    list_models=["Spherical", "Spherical"], empirical_variogram=df
)

xdem.spatialstats.plot_variogram(
    df,
    list_fit_fun=[func_sum_vgm1, func_sum_vgm2],
    list_fit_fun_label=["Single-range model", "Double-range model"],
    xscale_range_split=[100, 1000, 10000],
)

# %%
# The sum of two spherical models fits better, accouting for the small partial sill at longer ranges. Yet this longer
# range partial sill (correlated variance) is quite small...
#
# **So one could wonder: is it really important to account for this small additional "bump" in the variogram?**
#
# To answer this, we compute the precision of the DEM integrated over a certain surface area based on spatial integration of the
# variogram models using :func:`xdem.spatialstats.neff_circ`, with areas varying from pixel size to grid size.
# Numerical and exact integration of variogram is fast, allowing us to estimate errors for a wide range of areas rapidly.

areas = np.linspace(20, 10000, 50) ** 2

list_stderr_singlerange, list_stderr_doublerange, list_stderr_empirical = ([] for i in range(3))
for area in areas:

    # Number of effective samples integrated over the area for a single-range model
    neff_singlerange = xdem.spatialstats.number_effective_samples(area, params_vgm1)

    # For a double-range model
    neff_doublerange = xdem.spatialstats.number_effective_samples(area, params_vgm2)

    # Convert into a standard error
    stderr_singlerange = nmad(dh.data) / np.sqrt(neff_singlerange)
    stderr_doublerange = nmad(dh.data) / np.sqrt(neff_doublerange)
    list_stderr_singlerange.append(stderr_singlerange)
    list_stderr_doublerange.append(stderr_doublerange)

# %%
# We add an empirical error based on intensive Monte-Carlo sampling ("patches" method) to validate our results.
# This method is implemented in :func:`xdem.spatialstats.patches_method`. Here, we sample fewer areas to avoid for the
# patches method to run over long processing times, increasing from areas of 5 pixels to areas of 10000 pixels exponentially.

areas_emp = [4000 * 2 ** (i) for i in range(10)]
df_patches = xdem.spatialstats.patches_method(dh, gsd=dh.res[0], areas=areas_emp)


fig, ax = plt.subplots()
plt.plot(np.asarray(areas) / 1000000, list_stderr_singlerange, label="Single-range spherical model")
plt.plot(np.asarray(areas) / 1000000, list_stderr_doublerange, label="Double-range spherical model")
plt.scatter(
    df_patches.exact_areas.values / 1000000,
    df_patches.nmad.values,
    label="Empirical estimate",
    color="black",
    marker="x",
)
plt.xlabel("Averaging area (km²)")
plt.ylabel("Uncertainty in the mean elevation difference (m)")
plt.xscale("log")
plt.yscale("log")
plt.legend()
plt.show()

# %%
# *Note: in this example, we add a* ``random_state`` *argument to the patches method to yield a reproducible random
# sampling, and set* ``n_patches`` *to reduce computing time.*

# %%
# Using a single-range variogram highly underestimates the measurement error integrated over an area, by over a factor
# of ~100 for large surface areas. Using a double-range variogram brings us closer to the empirical error.
#
# **But, in this case, the error is still too small. Why?**
# The small size of the sampling area against the very large range of the noise implies that we might not verify the
# assumption of second-order stationarity (see :ref:`spatialstats`). Longer range correlations might be omitted by
# our analysis, due to the limits of the variogram sampling. In other words, a small part of the variance could be
# fully correlated over a large part of the grid: a vertical bias.
#
# As a first guess for this, let's examine the difference between mean and median to gain some insight on the central
# tendency of our sample:

diff_med_mean = np.nanmean(dh.data.data) - np.nanmedian(dh.data.data)
print(f"Difference mean/median: {diff_med_mean:.3f} meters.")

# %%
# If we now express it as a percentage of the dispersion:

print(f"{diff_med_mean/np.nanstd(dh.data)*100:.1f} % of STD.")

# %%
# There might be a significant bias of central tendency, i.e. almost fully correlated measurement error across the grid.
# If we assume that around 5% of the variance is fully correlated, and re-calculate our elevation measurement errors
# accordingly.

list_stderr_doublerange_plus_fullycorrelated = []
for area in areas:

    # For a double-range model
    neff_doublerange = xdem.spatialstats.neff_circular_approx_numerical(area=area, params_variogram_model=params_vgm2)

    # About 5% of the variance might be fully correlated, the other 95% has the random part that we quantified
    stderr_fullycorr = np.sqrt(0.05 * np.nanvar(dh.data))
    stderr_doublerange = np.sqrt(0.95 * np.nanvar(dh.data)) / np.sqrt(neff_doublerange)
    list_stderr_doublerange_plus_fullycorrelated.append(stderr_fullycorr + stderr_doublerange)

fig, ax = plt.subplots()
plt.plot(np.asarray(areas) / 1000000, list_stderr_singlerange, label="Single-range spherical model")
plt.plot(np.asarray(areas) / 1000000, list_stderr_doublerange, label="Double-range spherical model")
plt.plot(
    np.asarray(areas) / 1000000,
    list_stderr_doublerange_plus_fullycorrelated,
    label="5% fully correlated,\n 95% double-range spherical model",
)
plt.scatter(
    df_patches.exact_areas.values / 1000000,
    df_patches.nmad.values,
    label="Empirical estimate",
    color="black",
    marker="x",
)
plt.xlabel("Averaging area (km²)")
plt.ylabel("Uncertainty in the mean elevation difference (m)")
plt.xscale("log")
plt.yscale("log")
plt.legend()
plt.show()

# %%
# Our final estimation is now very close to the empirical error estimate.
#
# Some take-home points:
#   1. Long-range correlations are very important to reliably estimate measurement errors integrated in space, even if they have a small partial sill i.e. correlated variance,
#   2. Ideally, the grid must only contain correlation patterns significantly smaller than the grid size to verify second-order stationarity. Otherwise, be wary of small biases of central tendency, i.e. fully correlated measurement errors!



"""
Elevation error map
===================

Digital elevation models have a precision that can vary with terrain and instrument-related variables. Here, we
rely on a non-stationary spatial statistics framework to estimate and model this variability in elevation error,
using terrain slope and maximum curvature as explanatory variables, with stable terrain as an error proxy for moving
terrain.

**Reference**: `Hugonnet et al. (2022) <https://doi.org/10.1109/jstars.2022.3188922>`_, Figs. 4 and S6–S9. Equations 7
or 8 can be used to convert elevation change errors into elevation errors.
"""
import geoutils as gu

# sphinx_gallery_thumbnail_number = 1
import xdem

# %%
# We load a difference of DEMs at Longyearbyen, already coregistered using :ref:`coregistration-nuthkaab` as shown in
# the :ref:`sphx_glr_basic_examples_plot_nuth_kaab.py` example. We also load the reference DEM to derive terrain
# attributes and the glacier outlines here corresponding to moving terrain.
dh = xdem.DEM(xdem.examples.get_path("longyearbyen_ddem"))
ref_dem = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
glacier_outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

# %%
# We derive the terrain slope and maximum curvature from the reference DEM.
slope, maximum_curvature = xdem.terrain.get_terrain_attribute(ref_dem, attribute=["slope", "maximum_curvature"])

# %%
# Then, we run the pipeline for inference of elevation heteroscedasticity from stable terrain:
errors, df_binning, error_function = xdem.spatialstats.infer_heteroscedasticity_from_stable(
    dvalues=dh, list_var=[slope, maximum_curvature], list_var_names=["slope", "maxc"], unstable_mask=glacier_outlines
)

# %%
# The first output corresponds to the error map for the DEM (:math:`\pm` 1\ :math:`\sigma` level):
errors.show(vmin=2, vmax=7, cmap="Reds", cbar_title=r"Elevation error (1$\sigma$, m)")

# %%
# The second output is the dataframe of 2D binning with slope and maximum curvature:
df_binning

# %%
# The third output is the 2D binning interpolant, i.e. an error function with the slope and maximum curvature
# (*Note: below we divide the maximum curvature by 100 to convert it in* m\ :sup:`-1` ):
for slope, maxc in [(0, 0), (40, 0), (0, 5), (40, 5)]:
    print(
        "Error for a slope of {:.0f} degrees and"
        " {:.2f} m-1 max. curvature: {:.1f} m".format(slope, maxc / 100, error_function((slope, maxc)))
    )

# %%
# This pipeline will not always work optimally with default parameters: spread estimates can be affected by skewed
# distributions, the binning by extreme range of values, some DEMs do not have any error variability with terrain (e.g.,
# terrestrial photogrammetry). **To learn how to tune more parameters and use the subfunctions, see the gallery example:**
# :ref:`sphx_glr_advanced_examples_plot_heterosc_estimation_modelling.py`!



"""DEM class and functions."""
from __future__ import annotations

import pathlib
import warnings
from typing import Any, Literal

import rasterio as rio
from affine import Affine
from geoutils import SatelliteImage
from geoutils.raster import RasterType
from pyproj import CRS
from pyproj.crs import CompoundCRS, VerticalCRS

from xdem._typing import MArrayf, NDArrayf
from xdem.vcrs import (
    _build_ccrs_from_crs_and_vcrs,
    _grid_from_user_input,
    _parse_vcrs_name_from_product,
    _transform_zz,
    _vcrs_from_crs,
    _vcrs_from_user_input,
)

dem_attrs = ["_vcrs", "_vcrs_name", "_vcrs_grid"]


class DEM(SatelliteImage):  # type: ignore
    """
    The digital elevation model.

    The DEM has a single main attribute in addition to that inherited from :class:`geoutils.Raster`:
        vcrs: :class:`pyproj.VerticalCRS`
            Vertical coordinate reference system of the DEM.

    Other derivative attributes are:
        vcrs_name: :class:`str`
            Name of vertical CRS of the DEM.
        vcrs_grid: :class:`str`
            Grid path to the vertical CRS of the DEM.
        ccrs: :class:`pyproj.CompoundCRS`
            Compound vertical and horizontal CRS of the DEM.

    The attributes inherited from :class:`geoutils.Raster` are:
        data: :class:`np.ndarray`
            Data array of the DEM, with dimensions corresponding to (count, height, width).
        transform: :class:`affine.Affine`
            Geotransform of the DEM.
        crs: :class:`pyproj.crs.CRS`
            Coordinate reference system of the DEM.
        nodata: :class:`int` or :class:`float`
            Nodata value of the DEM.

    All other attributes are derivatives of those attributes, or read from the file on disk.
    See the API for more details.
    """

    def __init__(
        self,
        filename_or_dataset: str | RasterType | rio.io.DatasetReader | rio.io.MemoryFile,
        vcrs: Literal["Ellipsoid"]
        | Literal["EGM08"]
        | Literal["EGM96"]
        | VerticalCRS
        | str
        | pathlib.Path
        | int
        | None = None,
        silent: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Instantiate a digital elevation model.

        The vertical reference of the DEM can be defined by passing the `vcrs` argument.
        Otherwise, a vertical reference is tentatively parsed from the DEM product name.

        Inherits all attributes from the :class:`geoutils.Raster` and :class:`geoutils.SatelliteImage` classes.

        :param filename_or_dataset: The filename of the dataset.
        :param vcrs: Vertical coordinate reference system either as a name ("WGS84", "EGM08", "EGM96"),
            an EPSG code or pyproj.crs.VerticalCRS, or a path to a PROJ grid file (https://github.com/OSGeo/PROJ-data).
        :param silent: Whether to display vertical reference parsing.
        """

        self.data: NDArrayf
        self._vcrs: VerticalCRS | Literal["Ellipsoid"] | None = None
        self._vcrs_name: str | None = None
        self._vcrs_grid: str | None = None

        # If DEM is passed, simply point back to DEM
        if isinstance(filename_or_dataset, DEM):
            for key in filename_or_dataset.__dict__:
                setattr(self, key, filename_or_dataset.__dict__[key])
            return
        # Else rely on parent SatelliteImage class options (including raised errors)
        else:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Parse metadata from file not implemented")
                super().__init__(filename_or_dataset, silent=silent, **kwargs)

        # Ensure DEM has only one band: self.indexes can be None when data is not loaded through the Raster class
        if self.indexes is not None and len(self.indexes) > 1:
            raise ValueError("DEM rasters should be composed of one band only")

        # If the CRS in the raster metadata has a 3rd dimension, could set it as a vertical reference
        vcrs_from_crs = _vcrs_from_crs(CRS(self.crs))
        if vcrs_from_crs is not None:
            # If something was also provided by the user, user takes precedence
            # (we leave vcrs as it was for input)
            if vcrs is not None:
                # Raise a warning if the two are not the same
                vcrs_user = _vcrs_from_user_input(vcrs)
                if not vcrs_from_crs == vcrs_user:
                    warnings.warn(
                        "The CRS in the raster metadata already has a vertical component, "
                        "the user-input '{}' will override it.".format(vcrs)
                    )
            # Otherwise, use the one from the raster 3D CRS
            else:
                vcrs = vcrs_from_crs

        # If no vertical CRS was provided by the user or defined in the CRS
        if vcrs is None:
            vcrs = _parse_vcrs_name_from_product(self.product)

        # If a vertical reference was parsed or provided by user
        if vcrs is not None:
            self.set_vcrs(vcrs)

    def copy(self, new_array: NDArrayf | None = None) -> DEM:
        """
        Copy the DEM, possibly updating the data array.

        :param new_array: New data array.

        :return: Copied DEM.
        """

        new_dem = super().copy(new_array=new_array)  # type: ignore
        # The rest of attributes are immutable, including pyproj.CRS
        for attrs in dem_attrs:
            setattr(new_dem, attrs, getattr(self, attrs))

        return new_dem  # type: ignore

    @classmethod
    def from_array(
        cls: type[DEM],
        data: NDArrayf | MArrayf,
        transform: tuple[float, ...] | Affine,
        crs: CRS | int | None,
        nodata: int | float | None = None,
        vcrs: Literal["Ellipsoid"]
        | Literal["EGM08"]
        | Literal["EGM96"]
        | str
        | pathlib.Path
        | VerticalCRS
        | int
        | None = None,
    ) -> DEM:
        """Create a DEM from a numpy array and the georeferencing information.

        :param data: Input array.
        :param transform: Affine 2D transform. Either a tuple(x_res, 0.0, top_left_x,
            0.0, y_res, top_left_y) or an affine.Affine object.
        :param crs: Coordinate reference system. Either a rasterio CRS,
            or an EPSG integer.
        :param nodata: Nodata value.
        :param vcrs: Vertical coordinate reference system.

        :returns: DEM created from the provided array and georeferencing.
        """
        # We first apply the from_array of the parent class
        rast = SatelliteImage.from_array(data=data, transform=transform, crs=crs, nodata=nodata)
        # Then add the vcrs to the class call (that builds on top of the parent class)
        return cls(filename_or_dataset=rast, vcrs=vcrs)

    @property
    def vcrs(self) -> VerticalCRS | Literal["Ellipsoid"] | None:
        """Vertical coordinate reference system of the DEM."""

        return self._vcrs

    @property
    def vcrs_grid(self) -> str | None:
        """Grid path of vertical coordinate reference system of the DEM."""

        return self._vcrs_grid

    @property
    def vcrs_name(self) -> str | None:
        """Name of vertical coordinate reference system of the DEM."""

        if self.vcrs is not None:
            # If it is the ellipsoid
            if isinstance(self.vcrs, str):
                # Need to call CRS() here to make it work with rasterio.CRS...
                vcrs_name = f"Ellipsoid (No vertical CRS). Datum: {CRS(self.crs).ellipsoid.name}."
            # Otherwise, return the vertical reference name
            else:
                vcrs_name = self.vcrs.name
        else:
            vcrs_name = None

        return vcrs_name

    def set_vcrs(
        self,
        new_vcrs: Literal["Ellipsoid"] | Literal["EGM08"] | Literal["EGM96"] | str | pathlib.Path | VerticalCRS | int,
    ) -> None:
        """
        Set the vertical coordinate reference system of the DEM.

        :param new_vcrs: Vertical coordinate reference system either as a name ("Ellipsoid", "EGM08", "EGM96"),
            an EPSG code or pyproj.crs.VerticalCRS, or a path to a PROJ grid file (https://github.com/OSGeo/PROJ-data).
        """

        # Get vertical CRS and set it and the grid
        self._vcrs = _vcrs_from_user_input(vcrs_input=new_vcrs)
        self._vcrs_grid = _grid_from_user_input(vcrs_input=new_vcrs)

    @property
    def ccrs(self) -> CompoundCRS | CRS | None:
        """Compound horizontal and vertical coordinate reference system of the DEM."""

        if self.vcrs is not None:
            ccrs = _build_ccrs_from_crs_and_vcrs(crs=self.crs, vcrs=self.vcrs)
            return ccrs
        else:
            return None

    def to_vcrs(
        self,
        dst_vcrs: Literal["Ellipsoid", "EGM08", "EGM96"] | str | pathlib.Path | VerticalCRS | int,
        src_vcrs: Literal["Ellipsoid", "EGM08", "EGM96"] | str | pathlib.Path | VerticalCRS | int | None = None,
    ) -> None:
        """
        Convert the DEM to another vertical coordinate reference system.

        :param dst_vcrs: Destination vertical CRS. Either as a name ("WGS84", "EGM08", "EGM96"),
            an EPSG code or pyproj.crs.VerticalCRS, or a path to a PROJ grid file (https://github.com/OSGeo/PROJ-data)
        :param src_vcrs: Force a source vertical CRS (uses metadata by default). Same formats as for `dst_vcrs`.

        :return:
        """

        if self.vcrs is None and src_vcrs is None:
            raise ValueError(
                "The current DEM has no vertical reference, define one with .set_vref() "
                "or by passing `src_vcrs` to perform a conversion."
            )

        # Initial Compound CRS (only exists if vertical CRS is not None, as checked above)
        if src_vcrs is not None:
            # Warn if a vertical CRS already existed for that DEM
            if self.vcrs is not None:
                warnings.warn(
                    category=UserWarning,
                    message="Overriding the vertical CRS of the DEM with the one provided in `src_vcrs`.",
                )
            src_ccrs = _build_ccrs_from_crs_and_vcrs(self.crs, vcrs=src_vcrs)
        else:
            src_ccrs = self.ccrs

        # New destination Compound CRS
        dst_ccrs = _build_ccrs_from_crs_and_vcrs(self.crs, vcrs=_vcrs_from_user_input(vcrs_input=dst_vcrs))

        # If both compound CCRS are equal, do not run any transform
        if src_ccrs.equals(dst_ccrs):
            warnings.warn(
                message="Source and destination vertical CRS are the same, skipping vertical transformation.",
                category=UserWarning,
            )
            return None

        # Transform elevation with new vertical CRS
        zz = self.data
        xx, yy = self.coords(offset="center")
        zz_trans = _transform_zz(crs_from=src_ccrs, crs_to=dst_ccrs, xx=xx, yy=yy, zz=zz)

        # Update DEM
        self._data = zz_trans.astype(self.dtypes[0])  # type: ignore

        # Update vcrs (which will update ccrs if called)
        self.set_vcrs(new_vcrs=dst_vcrs)



"""
Normalized regional hypsometric interpolation
=============================================

There are many ways of interpolating gaps in a dDEM.
In the case of glaciers, one very useful fact is that elevation change generally varies with elevation.
This means that if valid pixels exist in a certain elevation bin, their values can be used to fill other pixels in the same approximate elevation.
Filling gaps by elevation is the main basis of "hypsometric interpolation approaches", of which there are many variations of.

One problem with simple hypsometric approaches is that they may not work for glaciers with different elevation ranges and scales.
Let's say we have two glaciers: one gigantic reaching from 0-1000 m, and one small from 900-1100 m.
Usually in the 2000s, glaciers thin rapidly at the bottom, while they may be neutral or only thin slightly in the top.
If we extrapolate the hypsometric signal of the gigantic glacier to use on the small one, it may seem like the smaller glacier has almost no change whatsoever.
This may be right, or it may be catastrophically wrong!

Normalized regional hypsometric interpolation solves the scale and elevation range problems in one go. It:

    1. Calculates a regional signal using the weighted average of each glacier's normalized signal:

        a. The glacier's elevation range is scaled from 0-1 to be elevation-independent.
        b. The glaciers elevation change is scaled from 0-1 to be magnitude-independent.
        c. A weight is assigned by the amount of valid pixels (well-covered large glaciers gain a higher weight)

    2. Re-scales that signal to fit each glacier once determined.

The consequence is a much more accurate interpolation approach that can be used in a multitude of glacierized settings.

"""
import geoutils as gu

# sphinx_gallery_thumbnail_number = 2
import matplotlib.pyplot as plt
import numpy as np

import xdem
import xdem.misc

# %%
# **Example files**

dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem_coreg"))

glacier_outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

# Rasterize the glacier outlines to create an index map.
# Stable ground is 0, the first glacier is 1, the second is 2, etc.
glacier_index_map = glacier_outlines.rasterize(dem_2009)

plt_extent = [
    dem_2009.bounds.left,
    dem_2009.bounds.right,
    dem_2009.bounds.bottom,
    dem_2009.bounds.top,
]


# %%
# To test the method, we can generate a semi-random mask to assign nans to glacierized areas.
# Let's remove 30% of the data.
np.random.seed(42)
random_nans = (xdem.misc.generate_random_field(dem_1990.shape, corr_size=5) > 0.7) & (glacier_index_map > 0)

random_nans.show()

# %%
# The normalized hypsometric signal shows the tendency for elevation change as a function of elevation.
# The magnitude may vary between glaciers, but the shape is generally similar.
# Normalizing by both elevation and elevation change, and then re-scaling the signal to every glacier, ensures that it is as accurate as possible.
# **NOTE**: The hypsometric signal does not need to be generated separately; it will be created by :func:`xdem.volume.norm_regional_hypsometric_interpolation`.
# Generating it first, however, allows us to visualize and validate it.

ddem = dem_2009 - dem_1990
ddem_voided = np.where(random_nans.data, np.nan, ddem.data)

signal = xdem.volume.get_regional_hypsometric_signal(
    ddem=ddem_voided,
    ref_dem=dem_2009.data,
    glacier_index_map=glacier_index_map,
)

plt.fill_between(signal.index.mid, signal["sigma-1-lower"], signal["sigma-1-upper"], label="Spread (+- 1 sigma)")
plt.plot(signal.index.mid, signal["w_mean"], color="black", label="Weighted mean")
plt.ylabel("Normalized elevation change")
plt.xlabel("Normalized elevation")
plt.legend()
plt.show()

# %%
# The signal can now be used (or simply estimated again if not provided) to interpolate the DEM.

ddem_filled = xdem.volume.norm_regional_hypsometric_interpolation(
    voided_ddem=ddem_voided, ref_dem=dem_2009, glacier_index_map=glacier_index_map, regional_signal=signal
)


plt.figure(figsize=(8, 5))
plt.imshow(ddem_filled.data, cmap="coolwarm_r", vmin=-10, vmax=10, extent=plt_extent)
plt.colorbar()
plt.show()


# %%
# We can plot the difference between the actual and the interpolated values, to validate the method.

difference = (ddem_filled - ddem)[random_nans]
median = np.nanmedian(difference)
nmad = xdem.spatialstats.nmad(difference)

plt.title(f"Median: {median:.2f} m, NMAD: {nmad:.2f} m")
plt.hist(difference.data, bins=np.linspace(-15, 15, 100))
plt.show()

# %%
# As we see, the median is close to zero, while the NMAD varies slightly more.
# This is expected, as the regional signal is good for multiple glaciers at once, but it cannot account for difficult local topography and meteorological conditions.
# It is therefore highly recommended for large regions; just don't zoom in too close!



"""
Estimation and modelling of heteroscedasticity
==============================================

Digital elevation models have a precision that can vary with terrain and instrument-related variables. This variability
in variance is called `heteroscedasticy <https://en.wikipedia.org/wiki/Homoscedasticity_and_heteroscedasticity>`_,
and rarely accounted for in DEM studies (see :ref:`intro`). Quantifying elevation heteroscedasticity is essential to
use stable terrain as an error proxy for moving terrain, and standardize data towards a stationary variance, necessary
to apply spatial statistics (see :ref:`spatialstats`).

Here, we show an advanced example in which we look for terrain-dependent explanatory variables to explain the
heteroscedasticity for a DEM difference at Longyearbyen. We use `data binning <https://en.wikipedia.org/wiki/Data_binning>`_
and robust statistics in N-dimension with :func:`xdem.spatialstats.nd_binning`, apply a N-dimensional interpolation with
:func:`xdem.spatialstats.interp_nd_binning`, and scale our interpolant function with a two-step standardization
:func:`xdem.spatialstats.two_step_standardization` to produce the final elevation error function.

**References**: `Hugonnet et al. (2021) <https://doi.org/10.1038/s41586-021-03436-z>`_, Equation 1, Extended Data Fig.
3a and `Hugonnet et al. (2022) <https://doi.org/10.1109/jstars.2022.3188922>`_, Figs. 4 and S6–S9. Equations 7 or 8 can
be used to convert elevation change errors into elevation errors.
"""
import geoutils as gu

# sphinx_gallery_thumbnail_number = 8
import matplotlib.pyplot as plt
import numpy as np

import xdem

# %%
# Here, we detail the steps used by ``xdem.spatialstats.infer_heteroscedasticity_from_stable`` exemplified in
# :ref:`sphx_glr_basic_examples_plot_infer_heterosc.py`. First, we load example files and create a glacier mask.

ref_dem = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
dh = xdem.DEM(xdem.examples.get_path("longyearbyen_ddem"))
glacier_outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
mask_glacier = glacier_outlines.create_mask(dh)

# %%
# We derive terrain attributes from the reference DEM (see :ref:`sphx_glr_basic_examples_plot_terrain_attributes.py`),
# which we will use to explore the variability in elevation error.
slope, aspect, planc, profc = xdem.terrain.get_terrain_attribute(
    dem=ref_dem, attribute=["slope", "aspect", "planform_curvature", "profile_curvature"]
)

# %%
# We convert to arrays and keep only stable terrain for the analysis of variability
dh_arr = dh[~mask_glacier].filled(np.nan)
slope_arr = slope[~mask_glacier].filled(np.nan)
aspect_arr = aspect[~mask_glacier].filled(np.nan)
planc_arr = planc[~mask_glacier].filled(np.nan)
profc_arr = profc[~mask_glacier].filled(np.nan)

# %%
# We use :func:`xdem.spatialstats.nd_binning` to perform N-dimensional binning on all those terrain variables, with uniform
# bin length divided by 30. We use the NMAD as a robust measure of `statistical dispersion <https://en.wikipedia.org/wiki/Statistical_dispersion>`_
# (see :ref:`robuststats-meanstd`).

df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[slope_arr, aspect_arr, planc_arr, profc_arr],
    list_var_names=["slope", "aspect", "planc", "profc"],
    statistics=["count", xdem.spatialstats.nmad],
    list_var_bins=30,
)

# %%
# We obtain a dataframe with the 1D binning results for each variable, the 2D binning results for all combinations of
# variables and the N-D (here 4D) binning with all variables.
# Overview of the dataframe structure for the 1D binning:
df[df.nd == 1]

# %%
# And for the 4D binning:
df[df.nd == 4]

# %%
# We can now visualize the results of the 1D binning of the computed NMAD of elevation differences with each variable
# using :func:`xdem.spatialstats.plot_1d_binning`.
# We can start with the slope that has been long known to be related to the elevation measurement error (e.g.,
# `Toutin (2002) <https://doi.org/10.1109/TGRS.2002.802878>`_).
xdem.spatialstats.plot_1d_binning(
    df, var_name="slope", statistic_name="nmad", label_var="Slope (degrees)", label_statistic="NMAD of dh (m)"
)

# %%
# We identify a clear variability, with the dispersion estimated from the NMAD increasing from ~2 meters for nearly flat
# slopes to above 12 meters for slopes steeper than 50°.
#
# What about the aspect?

xdem.spatialstats.plot_1d_binning(df, "aspect", "nmad", "Aspect (degrees)", "NMAD of dh (m)")

# %%
# There is no variability with the aspect that shows a dispersion averaging 2-3 meters, i.e. that of the complete sample.
#
# What about the plan curvature?

xdem.spatialstats.plot_1d_binning(df, "planc", "nmad", "Planform curvature (100 m$^{-1}$)", "NMAD of dh (m)")

# %%
# The relation with the plan curvature remains ambiguous.
# We should better define our bins to avoid sampling bins with too many or too few samples. For this, we can partition
# the data in quantiles in :func:`xdem.spatialstats.nd_binning`.
# *Note: we need a higher number of bins to work with quantiles and still resolve the edges of the distribution. As
# with many dimensions the ND bin size increases exponentially, we avoid binning all variables at the same
# time and instead bin one at a time.*
# We define 1000 quantile bins of size 0.001 (equivalent to 0.1% percentile bins) for the profile curvature:

df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[profc_arr],
    list_var_names=["profc"],
    statistics=["count", np.nanmedian, xdem.spatialstats.nmad],
    list_var_bins=[np.nanquantile(profc_arr, np.linspace(0, 1, 1000))],
)
xdem.spatialstats.plot_1d_binning(df, "profc", "nmad", "Profile curvature (100 m$^{-1}$)", "NMAD of dh (m)")

# %%
# We clearly identify a variability with the profile curvature, from 2 meters for low curvatures to above 4 meters
# for higher positive or negative curvature.
#
# What about the role of the plan curvature?

df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[planc_arr],
    list_var_names=["planc"],
    statistics=["count", np.nanmedian, xdem.spatialstats.nmad],
    list_var_bins=[np.nanquantile(planc_arr, np.linspace(0, 1, 1000))],
)
xdem.spatialstats.plot_1d_binning(df, "planc", "nmad", "Planform curvature (100 m$^{-1}$)", "NMAD of dh (m)")

# %%
# The plan curvature shows a similar relation. Those are symmetrical with 0, and almost equal for both types of curvature.
# To simplify the analysis, we here combine those curvatures into the maximum absolute curvature:

maxc_arr = np.maximum(np.abs(planc_arr), np.abs(profc_arr))
df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[maxc_arr],
    list_var_names=["maxc"],
    statistics=["count", np.nanmedian, xdem.spatialstats.nmad],
    list_var_bins=[np.nanquantile(maxc_arr, np.linspace(0, 1, 1000))],
)
xdem.spatialstats.plot_1d_binning(df, "maxc", "nmad", "Maximum absolute curvature (100 m$^{-1}$)", "NMAD of dh (m)")

# %%
# Here's our simplified relation! We now have both slope and maximum absolute curvature with clear variability of
# the elevation error.
#
# **But, one might wonder: high curvatures might occur more often around steep slopes than flat slope,
# so what if those two dependencies are actually one and the same?**
#
# We need to explore the variability with both slope and curvature at the same time:

df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[slope_arr, maxc_arr],
    list_var_names=["slope", "maxc"],
    statistics=["count", np.nanmedian, xdem.spatialstats.nmad],
    list_var_bins=30,
)

xdem.spatialstats.plot_2d_binning(
    df,
    var_name_1="slope",
    var_name_2="maxc",
    statistic_name="nmad",
    label_var_name_1="Slope (degrees)",
    label_var_name_2="Maximum absolute curvature (100 m$^{-1}$)",
    label_statistic="NMAD of dh (m)",
)

# %%
# We can see that part of the variability seems to be independent, but with the uniform bins it is hard to tell much
# more.
#
# If we use custom quantiles for both binning variables, and adjust the plot scale:

custom_bin_slope = np.unique(
    np.concatenate(
        [
            np.nanquantile(slope_arr, np.linspace(0, 0.95, 20)),
            np.nanquantile(slope_arr, np.linspace(0.96, 0.99, 5)),
            np.nanquantile(slope_arr, np.linspace(0.991, 1, 10)),
        ]
    )
)

custom_bin_curvature = np.unique(
    np.concatenate(
        [
            np.nanquantile(maxc_arr, np.linspace(0, 0.95, 20)),
            np.nanquantile(maxc_arr, np.linspace(0.96, 0.99, 5)),
            np.nanquantile(maxc_arr, np.linspace(0.991, 1, 10)),
        ]
    )
)

df = xdem.spatialstats.nd_binning(
    values=dh_arr,
    list_var=[slope_arr, maxc_arr],
    list_var_names=["slope", "maxc"],
    statistics=["count", np.nanmedian, xdem.spatialstats.nmad],
    list_var_bins=[custom_bin_slope, custom_bin_curvature],
)
xdem.spatialstats.plot_2d_binning(
    df,
    "slope",
    "maxc",
    "nmad",
    "Slope (degrees)",
    "Maximum absolute curvature (100 m$^{-1}$)",
    "NMAD of dh (m)",
    scale_var_2="log",
    vmin=2,
    vmax=10,
)


# %%
# We identify clearly that the two variables have an independent effect on the precision, with
#
# - *high curvatures and flat slopes* that have larger errors than *low curvatures and flat slopes*
# - *steep slopes and low curvatures* that have larger errors than *low curvatures and flat slopes* as well
#
# We also identify that, steep slopes (> 40°) only correspond to high curvature, while the opposite is not true, hence
# the importance of mapping the variability in two dimensions.
#
# Now we need to account for the heteroscedasticity identified. For this, the simplest approach is a numerical
# approximation i.e. a piecewise linear interpolation/extrapolation based on the binning results available through
# the function :func:`xdem.spatialstats.interp_nd_binning`. To ensure that only robust statistic values are used
# in the interpolation, we set a ``min_count`` value at 30 samples.

unscaled_dh_err_fun = xdem.spatialstats.interp_nd_binning(
    df, list_var_names=["slope", "maxc"], statistic="nmad", min_count=30
)

# %%
# The output is an interpolant function of slope and curvature that predicts the elevation error at any point. However,
# this predicted error might have a spread slightly off from that of the data:
#
# We compare the spread of the elevation difference on stable terrain and the average predicted error:
dh_err_stable = unscaled_dh_err_fun((slope_arr, maxc_arr))

print(
    "The spread of elevation difference is {:.2f} "
    "compared to a mean predicted elevation error of {:.2f}.".format(
        xdem.spatialstats.nmad(dh_arr), np.nanmean(dh_err_stable)
    )
)

# %%
# Thus, we rescale the function to exactly match the spread on stable terrain using the
# :func:`xdem.spatialstats.two_step_standardization` function, and get our final error function.

zscores, dh_err_fun = xdem.spatialstats.two_step_standardization(
    dh_arr, list_var=[slope_arr, maxc_arr], unscaled_error_fun=unscaled_dh_err_fun
)

for s, c in [(0.0, 0.1), (50.0, 0.1), (0.0, 20.0), (50.0, 20.0)]:
    print(
        "Elevation measurement error for slope of {:.0f} degrees, "
        "curvature of {:.2f} m-1: {:.1f}".format(s, c / 100, dh_err_fun((s, c))) + " meters."
    )

# %%
# This function can be used to estimate the spatial distribution of the elevation error on the extent of our DEMs:
maxc = np.maximum(np.abs(profc), np.abs(planc))
errors = dh.copy(new_array=dh_err_fun((slope.data, maxc.data)))

errors.show(cmap="Reds", vmin=2, vmax=8, cbar_title=r"Elevation error ($1\sigma$, m)")



"""Plot the comparison between a dDEM before and after Nuth and Kääb (2011) coregistration."""
import geoutils as gu
import matplotlib.pyplot as plt
import numpy as np

import xdem

dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
outlines_1990 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
inlier_mask = ~outlines_1990.create_mask(dem_2009)

nuth_kaab = xdem.coreg.NuthKaab()
nuth_kaab.fit(dem_2009, dem_1990, inlier_mask=inlier_mask)
dem_coreg = nuth_kaab.apply(dem_1990)

ddem_pre = dem_2009 - dem_1990
ddem_post = dem_2009 - dem_coreg

nmad_pre = xdem.spatialstats.nmad(ddem_pre[inlier_mask])
nmad_post = xdem.spatialstats.nmad(ddem_post[inlier_mask])

vlim = 20
plt.figure(figsize=(8, 5))
plt.subplot2grid((1, 15), (0, 0), colspan=7)
plt.title(f"Before coregistration. NMAD={nmad_pre:.1f} m")
plt.imshow(ddem_pre.data, cmap="coolwarm_r", vmin=-vlim, vmax=vlim)
plt.axis("off")
plt.subplot2grid((1, 15), (0, 7), colspan=7)
plt.title(f"After coregistration. NMAD={nmad_post:.1f} m")
img = plt.imshow(ddem_post.data, cmap="coolwarm_r", vmin=-vlim, vmax=vlim)
plt.axis("off")
plt.subplot2grid((1, 15), (0, 14), colspan=1)
cbar = plt.colorbar(img, fraction=0.4, ax=plt.gca())
cbar.set_label("Elevation change (m)")
plt.axis("off")

plt.tight_layout()
plt.show()



"""Plot an example of spatial interpolation of randomly generated errors."""
import geoutils as gu
import matplotlib.pyplot as plt
import numpy as np

import xdem

dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
outlines_1990 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

ddem = xdem.dDEM(dem_2009 - dem_1990, start_time=np.datetime64("1990-08-01"), end_time=np.datetime64("2009-08-01"))
# The example DEMs are void-free, so let's make some random voids.
ddem.data.mask = np.zeros_like(ddem.data, dtype=bool)  # Reset the mask
# Introduce 50000 nans randomly throughout the dDEM.
ddem.data.mask.ravel()[np.random.choice(ddem.data.size, 50000, replace=False)] = True

ddem.interpolate(method="linear")

ylim = (300, 100)
xlim = (800, 1050)

plt.figure(figsize=(8, 5))
plt.subplot(121)
plt.imshow(ddem.data.squeeze(), cmap="coolwarm_r", vmin=-50, vmax=50)
plt.ylim(ylim)
plt.xlim(xlim)
plt.axis("off")
plt.title("dDEM with random voids")
plt.subplot(122)
plt.imshow(ddem.filled_data.squeeze(), cmap="coolwarm_r", vmin=-50, vmax=50)
plt.ylim(ylim)
plt.xlim(xlim)
plt.axis("off")
plt.title("Linearly interpolated dDEM")


plt.tight_layout()
plt.show()



"""
Spatial correlation of errors
=============================

Digital elevation models have errors that are spatially correlated due to instrument or processing effects. Here, we
rely on a non-stationary spatial statistics framework to estimate and model spatial correlations in elevation error.
We use a sum of variogram forms to model this correlation, with stable terrain as an error proxy for moving terrain.

**Reference**: `Hugonnet et al. (2022) <https://doi.org/10.1109/jstars.2022.3188922>`_, Figure 5 and Equations 13–16.
"""
import geoutils as gu

# sphinx_gallery_thumbnail_number = 1
import xdem

# %%
# We load a difference of DEMs at Longyearbyen, already coregistered using :ref:`coregistration-nuthkaab` as shown in
# the :ref:`sphx_glr_basic_examples_plot_nuth_kaab.py` example. We also load the glacier outlines here corresponding to
# moving terrain.
dh = xdem.DEM(xdem.examples.get_path("longyearbyen_ddem"))
glacier_outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

# %%
# Then, we run the pipeline for inference of elevation heteroscedasticity from stable terrain (*Note: we pass a*
# ``random_state`` *argument to ensure a fixed, reproducible random subsampling in this example*). We ask for a fit with
# a Gaussian model for short range (as it is passed first), and Spherical for long range (as it is passed second):
(
    df_empirical_variogram,
    df_model_params,
    spatial_corr_function,
) = xdem.spatialstats.infer_spatial_correlation_from_stable(
    dvalues=dh, list_models=["Gaussian", "Spherical"], unstable_mask=glacier_outlines, random_state=42
)

# %%
# The first output corresponds to the dataframe of the empirical variogram, by default estimated using Dowd's estimator
# and the circular sampling scheme of ``skgstat.RasterEquidistantMetricSpace`` (Fig. S13 of Hugonnet et al. (2022)). The
# ``lags`` columns is the upper bound of spatial lag bins (lower bound of first bin being 0), the ``exp`` column is the
# "experimental" variance value of the variogram in that bin, the ``count`` the number of pairwise samples, and
# ``err_exp`` the 1-sigma error of the "experimental" variance, if more than one variogram is estimated with the
# ``n_variograms`` parameter.
df_empirical_variogram

# %%
# The second output is the dataframe of optimized model parameters (``range``, ``sill``, and possibly ``smoothness``)
# for a sum of gaussian and spherical models:
df_model_params

# %%
# The third output is the spatial correlation function with spatial lags, derived from the variogram:
for spatial_lag in [0, 100, 1000, 10000, 30000]:
    print(
        "Errors are correlated at {:.1f}% for a {:,.0f} m spatial lag".format(
            spatial_corr_function(spatial_lag) * 100, spatial_lag
        )
    )

# %%
# We can plot the empirical variogram and its model on a non-linear X-axis to identify the multi-scale correlations.
xdem.spatialstats.plot_variogram(
    df=df_empirical_variogram,
    list_fit_fun=[xdem.spatialstats.get_variogram_model_func(df_model_params)],
    xlabel="Spatial lag (m)",
    ylabel="Variance of\nelevation differences (m)",
    xscale_range_split=[100, 1000],
)

# %%
# This pipeline will not always work optimally with default parameters: variogram sampling is more robust with a lot of
# samples but takes long computing times, and the fitting might require multiple tries for forms and possibly bounds
# and first guesses to help the least-squares optimization. **To learn how to tune more parameters and use the
# subfunctions, see the gallery example:** :ref:`sphx_glr_advanced_examples_plot_variogram_estimation_modelling.py`!



"""Plot an example of regional hypsometric interpolation in central Svalbard."""
import geoutils as gu
import matplotlib.pyplot as plt
import numpy as np

import xdem

dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
outlines_1990 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

ddem = xdem.dDEM(dem_2009 - dem_1990, start_time=np.datetime64("1990-08-01"), end_time=np.datetime64("2009-08-01"))

ddem.data /= 2009 - 1990

mask = outlines_1990.create_mask(ddem)

ddem_bins = xdem.volume.hypsometric_binning(ddem[mask], dem_2009[mask])
stds = xdem.volume.hypsometric_binning(ddem[mask], dem_2009[mask], aggregation_function=np.std)

plt.figure(figsize=(8, 8))
plt.grid(zorder=0)


plt.plot(ddem_bins["value"], ddem_bins.index.mid, linestyle="--", zorder=1)

plt.barh(
    y=ddem_bins.index.mid,
    width=stds["value"],
    left=ddem_bins["value"] - stds["value"] / 2,
    height=(ddem_bins.index.left - ddem_bins.index.right) * 1,
    zorder=2,
    edgecolor="black",
)
for bin in ddem_bins.index:
    plt.vlines(ddem_bins.loc[bin, "value"], bin.left, bin.right, color="black", zorder=3)

plt.xlabel("Elevation change (m / a)")
plt.twiny()
plt.barh(
    y=ddem_bins.index.mid,
    width=ddem_bins["count"] / ddem_bins["count"].sum(),
    left=0,
    height=(ddem_bins.index.left - ddem_bins.index.right) * 1,
    zorder=2,
    alpha=0.2,
)
plt.xlabel("Normalized area distribution (hypsometry)")
plt.ylabel("Elevation (m a.s.l.)")

plt.tight_layout()
plt.show()



"""Functions to test the DEM collection tools."""
import datetime
import warnings

import geoutils as gu
import numpy as np

import xdem


class TestDEMCollection:
    dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
    dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    outlines_1990 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
    outlines_2010 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines_2010"))

    def test_init(self) -> None:

        timestamps = [datetime.datetime(1990, 8, 1), datetime.datetime(2009, 8, 1), datetime.datetime(2060, 8, 1)]

        scott_1990 = gu.Vector(self.outlines_1990.ds.loc[self.outlines_1990.ds["NAME"] == "Scott Turnerbreen"])
        scott_2010 = gu.Vector(self.outlines_2010.ds.loc[self.outlines_2010.ds["NAME"] == "Scott Turnerbreen"])

        # Make sure the glacier was bigger in 1990, since this is assumed later.
        assert scott_1990.ds.area.sum() > scott_2010.ds.area.sum()

        mask_2010 = scott_2010.create_mask(self.dem_2009)

        dem_2060 = self.dem_2009.copy()
        dem_2060[mask_2010] -= 30

        dems = xdem.DEMCollection(
            [self.dem_1990, self.dem_2009, dem_2060],
            timestamps=timestamps,
            outlines=dict(zip(timestamps[:2], [self.outlines_1990, self.outlines_2010])),
            reference_dem=1,
        )

        # Check that the first raster is the oldest one
        assert dems.dems[0].data.max() == self.dem_1990.data.max()
        assert dems.reference_dem.data.max() == self.dem_2009.data.max()

        dems.subtract_dems(resampling_method="nearest")

        assert np.mean(dems.ddems[0].data) < 0

        scott_filter = "NAME == 'Scott Turnerbreen'"

        dh_series = dems.get_dh_series(outlines_filter=scott_filter)

        # The 1990-2009 area should be the union of those years. The 2009-2060 area should just be the 2010 area.
        assert dh_series.iloc[0]["area"] > dh_series.iloc[-1]["area"]

        cumulative_dh = dems.get_cumulative_series(kind="dh", outlines_filter=scott_filter)
        cumulative_dv = dems.get_cumulative_series(kind="dv", outlines_filter=scott_filter)

        # Simple check that the cumulative_dh is overall negative.
        assert cumulative_dh.iloc[0] > cumulative_dh.iloc[-1]

        # Simple check that the dV number is of a greater magnitude than the dH number.
        assert abs(cumulative_dv.iloc[-1]) > abs(cumulative_dh.iloc[-1])

        # Generate 10000 NaN values randomly in one of the dDEMs
        dems.ddems[0].data[
            np.random.randint(0, dems.ddems[0].data.shape[0], 100),
            np.random.randint(0, dems.ddems[0].data.shape[1], 100),
        ] = np.nan
        # Check that the cumulative_dh function warns for NaNs
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            try:
                dems.get_cumulative_series(nans_ok=False)
            except UserWarning as exception:
                if "NaNs found in dDEM" not in str(exception):
                    raise exception

        # print(cumulative_dh)

        # raise NotImplementedError

    def test_dem_datetimes(self) -> None:
        """Try to create the DEMCollection without the timestamps argument (instead relying on datetime attributes)."""
        self.dem_1990.datetime = datetime.datetime(1990, 8, 1)
        self.dem_2009.datetime = datetime.datetime(2009, 8, 1)

        dems = xdem.DEMCollection([self.dem_1990, self.dem_2009])

        assert len(dems.timestamps) > 0

    def test_ddem_interpolation(self) -> None:
        """Test that dDEM interpolation works as it should."""
        # All warnings should raise errors from now on
        warnings.simplefilter("error")

        # Create a DEMCollection object
        dems = xdem.DEMCollection(
            [self.dem_2009, self.dem_1990], timestamps=[datetime.datetime(year, 8, 1) for year in (2009, 1990)]
        )

        # Create dDEMs
        dems.subtract_dems(resampling_method="nearest")

        # The example data does not have NaNs, so filled_data should exist.
        assert dems.ddems[0].filled_data is not None

        # Try to set the filled_data property with an invalid size.
        try:
            dems.ddems[0].filled_data = np.zeros(3)
        except AssertionError as exception:
            if "differs from the data shape" not in str(exception):
                raise exception

        # Generate 10000 NaN values randomly in one of the dDEMs
        dems.ddems[0].data[
            np.random.randint(0, dems.ddems[0].data.shape[0], 100),
            np.random.randint(0, dems.ddems[0].data.shape[1], 100),
        ] = np.nan

        # Make sure that filled_data is not available anymore, since the data now has nans
        assert dems.ddems[0].filled_data is None

        # Interpolate the nans
        dems.ddems[0].interpolate(method="linear")

        # Make sure that the filled_data is available again
        assert dems.ddems[0].filled_data is not None



"""
DEM coregistration classes and functions, including affine methods, bias corrections (i.e. non-affine) and filters.
"""

from xdem.coreg.affine import (  # noqa
    ICP,
    AffineCoreg,
    GradientDescending,
    NuthKaab,
    Tilt,
    VerticalShift,
)
from xdem.coreg.base import BlockwiseCoreg, Coreg, CoregPipeline, apply_matrix  # noqa
from xdem.coreg.biascorr import (  # noqa
    BiasCorr,
    BiasCorr1D,
    BiasCorr2D,
    BiasCorrND,
    Deramp,
    DirectionalBias,
    TerrainBias,
)
from xdem.coreg.workflows import dem_coregistration  # noqa



"""Functions to test the spatial statistics."""
from __future__ import annotations

import os
import warnings
from typing import Any

import geoutils as gu
import numpy as np
import pandas as pd
import pytest
import skgstat
from geoutils import Raster, Vector

import xdem
from xdem import examples
from xdem._typing import NDArrayf
from xdem.spatialstats import EmpiricalVariogramKArgs, nmad

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from skgstat import models

PLOT = False


def load_ref_and_diff() -> tuple[Raster, Raster, NDArrayf, Vector]:
    """Load example files to try coregistration methods with."""

    reference_raster = Raster(examples.get_path("longyearbyen_ref_dem"))
    outlines = Vector(examples.get_path("longyearbyen_glacier_outlines"))

    ddem = Raster(examples.get_path("longyearbyen_ddem"))
    mask = outlines.create_mask(ddem)

    return reference_raster, ddem, mask, outlines


class TestStats:

    # Load data for the entire test class
    ref, diff, mask, outlines = load_ref_and_diff()

    def test_nmad(self) -> None:
        """Test NMAD functionality runs on any type of input"""

        # Check that the NMAD is computed the same with a raster, masked array or NaN array
        nmad_raster = nmad(self.diff)
        nmad_ma = nmad(self.diff.data)
        nmad_array = nmad(self.diff.get_nanarray())

        assert nmad_raster == nmad_ma == nmad_array

        # Check that the scaling factor works
        nmad_1 = nmad(self.diff, nfact=1)
        nmad_2 = nmad(self.diff, nfact=2)

        assert nmad_1 * 2 == nmad_2


class TestBinning:

    # Load data for the entire test class
    ref, diff, mask, outlines = load_ref_and_diff()

    # Derive terrain attributes
    slope, aspect, maximum_curv = xdem.terrain.get_terrain_attribute(
        ref, attribute=["slope", "aspect", "maximum_curvature"]
    )

    def test_nd_binning(self) -> None:
        """Check that the nd_binning function works adequately and save dataframes to files for later tests"""

        # Subsampler
        indices = gu.raster.subsample_array(
            self.diff.data.flatten(), subsample=10000, return_indices=True, random_state=42
        )

        # 1D binning, by default will create 10 bins
        df = xdem.spatialstats.nd_binning(
            values=self.diff.data.flatten()[indices],
            list_var=[self.slope.data.flatten()[indices]],
            list_var_names=["slope"],
        )

        # Check length matches
        assert df.shape[0] == 10
        # Check bin edges match the minimum and maximum of binning variable
        assert np.nanmin(self.slope.data.flatten()[indices]) == np.min(pd.IntervalIndex(df.slope).left)
        assert np.nanmax(self.slope.data.flatten()[indices]) == np.max(pd.IntervalIndex(df.slope).right)

        # NMAD should go up quite a bit with slope, more than 8 m between the two extreme bins
        assert df.nmad.values[-1] - df.nmad.values[0] > 8

        # 1D binning with 20 bins
        df = xdem.spatialstats.nd_binning(
            values=self.diff.data.flatten()[indices],
            list_var=[self.slope.data.flatten()[indices]],
            list_var_names=["slope"],
            list_var_bins=20,
        )
        # Check length matches
        assert df.shape[0] == 20

        # Define function for custom stat
        def percentile_80(a: NDArrayf) -> np.floating[Any]:
            return np.nanpercentile(a, 80)

        # Check the function runs with custom functions
        df = xdem.spatialstats.nd_binning(
            values=self.diff.data.flatten()[indices],
            list_var=[self.slope.data.flatten()[indices]],
            list_var_names=["slope"],
            statistics=[percentile_80],
        )
        # Check that the count is added automatically by the function when not user-defined
        assert "count" in df.columns.values

        # 2D binning
        df = xdem.spatialstats.nd_binning(
            values=self.diff.data.flatten()[indices],
            list_var=[self.slope.data.flatten()[indices], self.ref.data.flatten()[indices]],
            list_var_names=["slope", "elevation"],
        )

        # Dataframe should contain two 1D binning of length 10 and one 2D binning of length 100
        assert df.shape[0] == (10 + 10 + 100)

        # 3D binning
        df = xdem.spatialstats.nd_binning(
            values=self.diff.data.flatten()[indices],
            list_var=[
                self.slope.data.flatten()[indices],
                self.ref.data.flatten()[indices],
                self.aspect.data.flatten()[indices],
            ],
            list_var_names=["slope", "elevation", "aspect"],
            list_var_bins=4,
        )

        # Dataframe should contain three 1D binning of length 10 and three 2D binning of length 100 and one 2D binning
        # of length 1000
        assert df.shape[0] == (4**3 + 3 * 4**2 + 3 * 4)

        # Save for later use
        df.to_csv(os.path.join(examples._EXAMPLES_DIRECTORY, "df_3d_binning_slope_elevation_aspect.csv"), index=False)

    def test_interp_nd_binning_artificial_data(self) -> None:
        """Check that the N-dimensional interpolation works correctly using artificial data"""

        # Check the function works with a classic input (see example)
        df = pd.DataFrame(
            {
                "var1": [1, 2, 3, 1, 2, 3, 1, 2, 3],
                "var2": [1, 1, 1, 2, 2, 2, 3, 3, 3],
                "statistic": [1, 2, 3, 4, 5, 6, 7, 8, 9],
            }
        )
        arr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]).reshape((3, 3))
        fun = xdem.spatialstats.interp_nd_binning(
            df, list_var_names=["var1", "var2"], statistic="statistic", min_count=None
        )

        # Check that the dimensions are rightly ordered
        assert fun((1, 3)) == df[np.logical_and(df["var1"] == 1, df["var2"] == 3)]["statistic"].values[0]
        assert fun((3, 1)) == df[np.logical_and(df["var1"] == 3, df["var2"] == 1)]["statistic"].values[0]

        # Check interpolation falls right on values for points (1, 1), (1, 2) etc...
        for i in range(3):
            for j in range(3):
                x = df["var1"][3 * i + j]
                y = df["var2"][3 * i + j]
                stat = df["statistic"][3 * i + j]
                assert fun((x, y)) == stat

        # Check bilinear interpolation inside the grid
        points_in = [(1.5, 1.5), (1.5, 2.5), (2.5, 1.5), (2.5, 2.5)]
        for point in points_in:
            # The values are 1 off from Python indexes
            x = point[0] - 1
            y = point[1] - 1
            # Get four closest points on the grid
            xlow = int(x - 0.5)
            xupp = int(x + 0.5)
            ylow = int(y - 0.5)
            yupp = int(y + 0.5)
            # Check the bilinear interpolation matches the mean value of those 4 points (equivalent as its the middle)
            assert fun((y + 1, x + 1)) == np.mean([arr[xlow, ylow], arr[xupp, ylow], arr[xupp, yupp], arr[xlow, yupp]])

        # Check bilinear extrapolation for points at 1 spacing outside from the input grid
        points_out = (
            [(0, i) for i in np.arange(1, 4)]
            + [(i, 0) for i in np.arange(1, 4)]
            + [(4, i) for i in np.arange(1, 4)]
            + [(i, 4) for i in np.arange(4, 1)]
        )
        for point in points_out:
            x = point[0] - 1
            y = point[1] - 1
            val_extra = fun((y + 1, x + 1))
            # (OUTDATED: The difference between the points extrapolated outside should be linear with the grid edges,
            # # i.e. the same as the difference as the first points inside the grid along the same axis)
            # if point[0] == 0:
            #     diff_in = arr[x + 2, y] - arr[x + 1, y]
            #     diff_out = arr[x + 1, y] - val_extra
            # elif point[0] == 4:
            #     diff_in = arr[x - 2, y] - arr[x - 1, y]
            #     diff_out = arr[x - 1, y] - val_extra
            # elif point[1] == 0:
            #     diff_in = arr[x, y + 2] - arr[x, y + 1]
            #     diff_out = arr[x, y + 1] - val_extra
            # # has to be y == 4
            # else:
            #     diff_in = arr[x, y - 2] - arr[x, y - 1]
            #     diff_out = arr[x, y - 1] - val_extra
            # assert diff_in == diff_out

            # Update with nearest default: the value should be that of the nearest!
            if point[0] == 0:
                near = arr[x + 1, y]
            elif point[0] == 4:
                near = arr[x - 1, y]
            elif point[1] == 0:
                near = arr[x, y + 1]
            else:
                near = arr[x, y - 1]
            assert near == val_extra

        # Check that the output extrapolates as "nearest neighbour" far outside the grid
        points_far_out = (
            [(-10, i) for i in np.arange(1, 4)]
            + [(i, -10) for i in np.arange(1, 4)]
            + [(14, i) for i in np.arange(1, 4)]
            + [(i, 14) for i in np.arange(4, 1)]
        )
        for point in points_far_out:
            x = point[0] - 1
            y = point[1] - 1
            val_extra = fun((y + 1, x + 1))
            # Update with nearest default: the value should be that of the nearest!
            if point[0] == -10:
                near = arr[0, y]
            elif point[0] == 14:
                near = arr[-1, y]
            elif point[1] == -10:
                near = arr[x, 0]
            else:
                near = arr[x, -1]
            assert near == val_extra

        # Check that the output is rightly ordered in 3 dimensions, and works with varying dimension lengths
        vec1 = np.arange(1, 3)
        vec2 = np.arange(1, 4)
        vec3 = np.arange(1, 5)
        x, y, z = np.meshgrid(vec1, vec2, vec3)
        df = pd.DataFrame(
            {"var1": x.ravel(), "var2": y.ravel(), "var3": z.ravel(), "statistic": np.arange(len(x.ravel()))}
        )
        fun = xdem.spatialstats.interp_nd_binning(
            df, list_var_names=["var1", "var2", "var3"], statistic="statistic", min_count=None
        )
        for i in vec1:
            for j in vec2:
                for k in vec3:
                    assert (
                        fun((i, j, k))
                        == df[np.logical_and.reduce((df["var1"] == i, df["var2"] == j, df["var3"] == k))][
                            "statistic"
                        ].values[0]
                    )

    def test_interp_nd_binning_realdata(self) -> None:
        """Check that the function works well with outputs from the nd_binning function"""

        # Read nd_binning output
        df = pd.read_csv(
            os.path.join(examples._EXAMPLES_DIRECTORY, "df_3d_binning_slope_elevation_aspect.csv"), index_col=None
        )

        # First, in 1D
        fun = xdem.spatialstats.interp_nd_binning(df, list_var_names="slope")

        # Check a value is returned inside the grid
        assert np.isfinite(fun(([15],)))
        # Check the nmad increases with slope
        assert fun(([20],)) > fun(([0],))
        # Check a value is returned outside the grid
        assert all(np.isfinite(fun(([-5, 50],))))

        # Check when the first passed binning variable contains NaNs because of other binning variable
        fun = xdem.spatialstats.interp_nd_binning(df, list_var_names="elevation")

        # Then, in 2D
        fun = xdem.spatialstats.interp_nd_binning(df, list_var_names=["slope", "elevation"])

        # Check a value is returned inside the grid
        assert np.isfinite(fun(([15], [1000])))
        # Check the nmad increases with slope
        assert fun(([40], [300])) > fun(([10], [300]))
        # Check a value is returned outside the grid
        assert all(np.isfinite(fun(([-5, 50], [-500, 3000]))))

        # Then in 3D
        fun = xdem.spatialstats.interp_nd_binning(df, list_var_names=["slope", "elevation", "aspect"])

        # Check a value is returned inside the grid
        assert np.isfinite(fun(([15], [1000], [np.pi])))
        # Check the nmad increases with slope
        assert fun(([30], [300], [np.pi])) > fun(([10], [300], [np.pi]))
        # Check a value is returned outside the grid
        assert all(np.isfinite(fun(([-5, 50], [-500, 3000], [-2 * np.pi, 4 * np.pi]))))

    def test_get_perbin_nd_binning(self) -> None:
        """Test the get per-bin function."""

        # Read nd_binning output
        df = pd.read_csv(
            os.path.join(examples._EXAMPLES_DIRECTORY, "df_3d_binning_slope_elevation_aspect.csv"), index_col=None
        )

        # Get values for arrays from the above 3D binning
        perbin_values = xdem.spatialstats.get_perbin_nd_binning(
            df=df,
            list_var=[
                self.slope.data,
                self.ref.data,
                self.aspect.data,
            ],
            list_var_names=["slope", "elevation", "aspect"],
        )

        # Check that the function preserves the shape
        assert np.shape(self.slope.data) == np.shape(perbin_values)

        # Check that the bin are rightly recognized
        df = df[df.nd == 3]
        # Convert the intervals from string due to saving to file
        for var in ["slope", "elevation", "aspect"]:
            df[var] = [xdem.spatialstats._pandas_str_to_interval(x) for x in df[var]]

        # Take 1000 random points in the array
        np.random.seed(42)
        xrand = np.random.randint(low=0, high=perbin_values.shape[0], size=1000)
        yrand = np.random.randint(low=0, high=perbin_values.shape[1], size=1000)

        for i in range(len(xrand)):

            # Get the value at the random point for elevation, slope, aspect
            x = xrand[i]
            y = yrand[i]
            h = self.ref.data[x, y]
            slp = self.slope.data[x, y]
            asp = self.aspect.data[x, y]

            if np.logical_or.reduce((np.isnan(h), np.isnan(slp), np.isnan(asp))):
                continue

            # Isolate the bin in the dataframe, should be only one
            index_bin = np.logical_and.reduce(
                (
                    [h in interv for interv in df["elevation"]],
                    [slp in interv for interv in df["slope"]],
                    [asp in interv for interv in df["aspect"]],
                )
            )
            assert np.count_nonzero(index_bin) == 1

            # Get the statistic value and verify that this was the one returned by the function
            statistic_value = df["nanmedian"][index_bin].values[0]
            # Nan equality does not work, so we compare finite values first
            if ~np.isnan(statistic_value):
                assert statistic_value == perbin_values[x, y]
            # And then check that a NaN is returned if it is the statistic
            else:
                assert np.isnan(perbin_values[x, y])

    def test_two_step_standardization(self) -> None:
        """Test two-step standardization function"""

        # Reproduce the first steps of binning
        df_binning = xdem.spatialstats.nd_binning(
            values=self.diff[~self.mask],
            list_var=[self.slope[~self.mask], self.maximum_curv[~self.mask]],
            list_var_names=["var1", "var2"],
            statistics=[xdem.spatialstats.nmad],
        )
        unscaled_fun = xdem.spatialstats.interp_nd_binning(
            df_binning, list_var_names=["var1", "var2"], statistic="nmad"
        )
        # The zscore spread should not be one right after binning
        zscores = self.diff[~self.mask] / unscaled_fun((self.slope[~self.mask], self.maximum_curv[~self.mask]))
        scale_fac = xdem.spatialstats.nmad(zscores)
        assert scale_fac != 1

        # Filter with a factor of 3 and the standard deviation (not default values) and check the function outputs
        # the exact same array
        zscores[np.abs(zscores) > 3 * np.nanstd(zscores)] = np.nan
        scale_fac_std = np.nanstd(zscores)
        zscores /= scale_fac_std
        zscores_2, final_func = xdem.spatialstats.two_step_standardization(
            dvalues=self.diff[~self.mask],
            list_var=[self.slope[~self.mask], self.maximum_curv[~self.mask]],
            unscaled_error_fun=unscaled_fun,
            spread_statistic=np.nanstd,
            fac_spread_outliers=3,
        )
        assert np.array_equal(zscores, zscores_2, equal_nan=True)

        # Check the output of the scaled function is simply the unscaled one times the spread statistic
        test_slopes = np.linspace(0, 50, 50)
        test_max_curvs = np.linspace(0, 10, 50)
        assert np.array_equal(
            unscaled_fun((test_slopes, test_max_curvs)) * scale_fac_std, final_func((test_slopes, test_max_curvs))
        )

    def test_estimate_model_heteroscedasticity_and_infer_from_stable(self) -> None:
        """Test consistency of outputs and errors in wrapper functions for estimation of heteroscedasticity"""

        # Test infer function
        errors_1, df_binning_1, err_fun_1 = xdem.spatialstats.infer_heteroscedasticity_from_stable(
            dvalues=self.diff, list_var=[self.slope, self.maximum_curv], unstable_mask=self.outlines
        )

        df_binning_2, err_fun_2 = xdem.spatialstats.estimate_model_heteroscedasticity(
            dvalues=self.diff[~self.mask],
            list_var=[self.slope[~self.mask], self.maximum_curv[~self.mask]],
            list_var_names=["var1", "var2"],
        )

        pd.testing.assert_frame_equal(df_binning_1, df_binning_2)
        test_slopes = np.linspace(0, 50, 50)
        test_max_curvs = np.linspace(0, 10, 50)
        assert np.array_equal(err_fun_1((test_slopes, test_max_curvs)), err_fun_2((test_slopes, test_max_curvs)))

        # Test the error map is consistent as well
        errors_2_arr = err_fun_2((self.slope.get_nanarray(), self.maximum_curv.get_nanarray()))
        errors_1_arr = gu.raster.get_array_and_mask(errors_1)[0]
        assert np.array_equal(errors_1_arr, errors_2_arr, equal_nan=True)

        # Save for use in TestVariogram
        errors_1.save(os.path.join(examples._EXAMPLES_DIRECTORY, "dh_error.tif"))

        # Check that errors are raised with wrong input
        with pytest.raises(ValueError, match="The values must be a Raster or NumPy array, or a list of those."):
            xdem.spatialstats.infer_heteroscedasticity_from_stable(
                dvalues="not_an_array", stable_mask=~self.mask, list_var=[self.slope.get_nanarray()]
            )
        with pytest.raises(ValueError, match="The stable mask must be a Vector, Mask, GeoDataFrame or NumPy array."):
            xdem.spatialstats.infer_heteroscedasticity_from_stable(
                dvalues=self.diff, stable_mask="not_a_vector_or_array", list_var=[self.slope.get_nanarray()]
            )
        with pytest.raises(ValueError, match="The unstable mask must be a Vector, Mask, GeoDataFrame or NumPy array."):
            xdem.spatialstats.infer_heteroscedasticity_from_stable(
                dvalues=self.diff, unstable_mask="not_a_vector_or_array", list_var=[self.slope.get_nanarray()]
            )

        with pytest.raises(
            ValueError,
            match="The stable mask can only passed as a Vector or GeoDataFrame if the input "
            "values contain a Raster.",
        ):
            xdem.spatialstats.infer_heteroscedasticity_from_stable(
                dvalues=self.diff.get_nanarray(), stable_mask=self.outlines, list_var=[self.slope.get_nanarray()]
            )

    def test_plot_binning(self) -> None:

        # Define placeholder data
        df = pd.DataFrame({"var1": [0, 1, 2], "var2": [2, 3, 4], "statistic": [0, 0, 0]})

        # Check that the 1D plotting fails with a warning if the variable or statistic is not well-defined
        with pytest.raises(ValueError, match='The variable "var3" is not part of the provided dataframe column names.'):
            xdem.spatialstats.plot_1d_binning(df, var_name="var3", statistic_name="statistic")
        with pytest.raises(
            ValueError, match='The statistic "stat" is not part of the provided dataframe column names.'
        ):
            xdem.spatialstats.plot_1d_binning(df, var_name="var1", statistic_name="stat")

        # Same for the 2D plotting
        with pytest.raises(ValueError, match='The variable "var3" is not part of the provided dataframe column names.'):
            xdem.spatialstats.plot_2d_binning(df, var_name_1="var3", var_name_2="var1", statistic_name="statistic")
        with pytest.raises(
            ValueError, match='The statistic "stat" is not part of the provided dataframe column names.'
        ):
            xdem.spatialstats.plot_2d_binning(df, var_name_1="var1", var_name_2="var1", statistic_name="stat")


class TestVariogram:

    ref, diff, mask, outlines = load_ref_and_diff()

    def test_sample_multirange_variogram_default(self) -> None:
        """Verify that the default function runs, and its basic output"""

        # Check the variogram output is consistent for a random state
        df = xdem.spatialstats.sample_empirical_variogram(values=self.diff, subsample=10, random_state=42)
        assert df["exp"][15] == pytest.approx(5.046060943603516, abs=1e-3)
        assert df["lags"][15] == pytest.approx(5120)
        assert df["count"][15] == 5
        # With a single run, no error can be estimated
        assert all(np.isnan(df.err_exp.values))

        # Check that all type of coordinate inputs work
        # Only the array and the ground sampling distance
        xdem.spatialstats.sample_empirical_variogram(
            values=self.diff.data, gsd=self.diff.res[0], subsample=10, random_state=42, verbose=True
        )

        # Test multiple runs
        df2 = xdem.spatialstats.sample_empirical_variogram(
            values=self.diff, subsample=10, random_state=42, n_variograms=2
        )

        # Check that an error is estimated
        assert any(~np.isnan(df2.err_exp.values))

        # Test that running on several cores does not trigger any error
        xdem.spatialstats.sample_empirical_variogram(
            values=self.diff, subsample=10, random_state=42, n_variograms=2, n_jobs=2
        )

        # Test plotting of empirical variogram by itself
        if PLOT:
            xdem.spatialstats.plot_variogram(df2)

    def test_sample_empirical_variogram_speed(self) -> None:
        """Verify that no speed is lost outside of routines on variogram sampling by comparing manually to skgstat"""

        values = self.diff
        subsample = 10

        # First, run the xDEM wrapper function
        # t0 = time.time()
        df = xdem.spatialstats.sample_empirical_variogram(values=values, subsample=subsample, random_state=42)
        # t1 = time.time()

        # Second, do it manually with skgstat

        # Ground sampling distance
        gsd = values.res[0]
        # Coords
        nx, ny = values.shape
        x, y = np.meshgrid(np.arange(0, values.shape[0] * gsd, gsd), np.arange(0, values.shape[1] * gsd, gsd))
        coords = np.dstack((x.ravel(), y.ravel())).squeeze()

        # Redefine parameters fed to skgstat manually
        # Maxlag
        maxlag = np.sqrt(
            (np.max(coords[:, 0]) - np.min(coords[:, 0])) ** 2 + (np.max(coords[:, 1]) - np.min(coords[:, 1])) ** 2
        )

        # Binning function
        bin_func = []
        right_bin_edge = np.sqrt(2) * gsd
        while right_bin_edge < maxlag:
            bin_func.append(right_bin_edge)
            # We use the default exponential increasing factor of RasterEquidistantMetricSpace, adapted for grids
            right_bin_edge *= np.sqrt(2)
        bin_func.append(maxlag)
        # Extent
        extent = (np.min(coords[:, 0]), np.max(coords[:, 0]), np.min(coords[:, 1]), np.max(coords[:, 1]))
        # Shape
        shape = values.shape
        # Random state
        rnd = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(42)))

        keyword_arguments = {"subsample": subsample, "extent": extent, "shape": shape, "verbose": False}
        runs, samples, ratio_subsample = xdem.spatialstats._choose_cdist_equidistant_sampling_parameters(
            **keyword_arguments
        )

        # Index of valid values
        values_arr, mask_nodata = gu.raster.get_array_and_mask(values)

        # t3 = time.time()
        rems = skgstat.RasterEquidistantMetricSpace(
            coords=coords[~mask_nodata.ravel(), :],
            shape=shape,
            extent=extent,
            samples=samples,
            ratio_subsample=ratio_subsample,
            runs=runs,
            rnd=rnd,
        )
        V = skgstat.Variogram(
            rems,
            values=values_arr[~mask_nodata].ravel(),
            normalize=False,
            fit_method=None,
            bin_func=bin_func,
            maxlag=maxlag,
        )
        # t4 = time.time()

        # Get bins, empirical variogram values, and bin count
        bins, exp = V.get_empirical(bin_center=False)
        count = V.bin_count

        # Write to dataframe
        df2 = pd.DataFrame()
        df2 = df2.assign(exp=exp, bins=bins, count=count)
        df2 = df2.rename(columns={"bins": "lags"})
        df2["err_exp"] = np.nan
        df2.drop(df2.tail(1).index, inplace=True)
        df2 = df2.astype({"exp": "float64", "err_exp": "float64", "lags": "float64", "count": "int64"})

        # t2 = time.time()

        # Check if the two frames are equal
        pd.testing.assert_frame_equal(df, df2)

        # Check that the two ways are taking the same time with 50% margin
        # time_method_1 = t1 - t0
        # time_method_2 = t2 - t1
        # assert time_method_1 == pytest.approx(time_method_2, rel=0.5)

        # Check that all this time is based on variogram sampling at about 70%, even with the smallest number of
        # samples of 10
        # time_metricspace_variogram = t4 - t3
        # assert time_metricspace_variogram == pytest.approx(time_method_2, rel=0.3)

    @pytest.mark.parametrize(
        "subsample_method", ["pdist_point", "pdist_ring", "pdist_disk", "cdist_point"]
    )  # type: ignore
    def test_sample_multirange_variogram_methods(self, subsample_method) -> None:
        """Verify that all other methods run"""

        # Check the variogram estimation runs for several methods
        df = xdem.spatialstats.sample_empirical_variogram(
            values=self.diff, subsample=10, random_state=42, subsample_method=subsample_method
        )

        assert not df.empty

        # Check that the output is correct
        expected_columns = ["exp", "lags", "count"]
        expected_dtypes = [np.float64, np.float64, np.int64]
        for col in expected_columns:
            # Check that the column exists
            assert col in df.columns
            # Check that the column has the correct dtype
            assert df[col].dtype == expected_dtypes[expected_columns.index(col)]

    def test_sample_multirange_variogram_args(self) -> None:
        """Verify that optional parameters run only for their specific method, raise warning otherwise"""

        # Define parameters
        pdist_args: EmpiricalVariogramKArgs = {"pdist_multi_ranges": [0, self.diff.res[0] * 5, self.diff.res[0] * 10]}
        cdist_args: EmpiricalVariogramKArgs = {"ratio_subsample": 0.5, "runs": 10}
        nonsense_args = {"thisarg": "shouldnotexist"}

        # Check the function raises a warning for optional arguments incorrect to the method
        with pytest.warns(UserWarning):
            # An argument only use by cdist with a pdist method
            xdem.spatialstats.sample_empirical_variogram(
                values=self.diff, subsample=10, random_state=42, subsample_method="pdist_ring", **cdist_args
            )

        with pytest.warns(UserWarning):
            # Same here
            xdem.spatialstats.sample_empirical_variogram(
                values=self.diff,
                subsample=10,
                random_state=42,
                subsample_method="cdist_equidistant",
                **pdist_args,
            )

        with pytest.warns(UserWarning):
            # Should also raise a warning for a nonsense argument
            xdem.spatialstats.sample_empirical_variogram(
                values=self.diff,
                subsample=10,
                random_state=42,
                subsample_method="cdist_equidistant",
                **nonsense_args,  # type: ignore
            )

        # Check the function passes optional arguments specific to pdist methods without warning
        xdem.spatialstats.sample_empirical_variogram(
            values=self.diff, subsample=10, random_state=42, subsample_method="pdist_ring", **pdist_args
        )

        # Check the function passes optional arguments specific to cdist methods without warning
        xdem.spatialstats.sample_empirical_variogram(
            values=self.diff, random_state=42, subsample=10, subsample_method="cdist_equidistant", **cdist_args
        )

    # N is the number of samples in an ensemble
    @pytest.mark.parametrize("subsample", [10, 100, 1000, 10000])  # type: ignore
    @pytest.mark.parametrize("shape", [(50, 50), (100, 100), (500, 500)])  # type: ignore
    def test_choose_cdist_equidistant_sampling_parameters(self, subsample: int, shape: tuple[int, int]) -> None:
        """Verify that the automatically-derived parameters of equidistant sampling are sound"""

        # Assign an arbitrary extent
        extent = (0, 1, 0, 1)

        # The number of different pairwise combinations in a single ensemble (scipy.pdist function) is N*(N-1)/2
        # which is approximately N**2/2
        pdist_pairwise_combinations = subsample**2 / 2

        # Run the function
        keyword_arguments = {"subsample": subsample, "extent": extent, "shape": shape, "verbose": False}
        runs, samples, ratio_subsample = xdem.spatialstats._choose_cdist_equidistant_sampling_parameters(
            **keyword_arguments
        )

        # There is at least 2 samples
        assert samples > 2
        # Can only be maximum 100 runs
        assert runs <= 100

        # Get maxdist
        maxdist = np.sqrt((extent[1] - extent[0]) ** 2 + (extent[3] - extent[2]) ** 2)
        res = np.mean([(extent[1] - extent[0]) / (shape[0] - 1), (extent[3] - extent[2]) / (shape[1] - 1)])
        # Then, we compute the radius from the center ensemble with the default value of subsample ratio in the function
        # skgstat.RasterEquidistantMetricSpace
        center_radius = np.sqrt(1.0 / ratio_subsample * samples / np.pi) * res
        nb_rings_final = int(2 * np.log(maxdist / center_radius) / np.log(2))
        cdist_pairwise_combinations = runs * samples**2 * nb_rings_final

        # Check the number of pairwise comparisons are the same (within 50%, due to rounding as integers)
        assert pdist_pairwise_combinations == pytest.approx(cdist_pairwise_combinations, rel=0.5, abs=10)

    def test_errors_subsample_parameter(self) -> None:
        """Tests that an error is raised when the subsample argument is too little"""

        keyword_arguments = {"subsample": 3, "extent": (0, 1, 0, 1), "shape": (10, 10), "verbose": False}

        with pytest.raises(ValueError, match="The number of subsamples needs to be at least 10."):
            xdem.spatialstats._choose_cdist_equidistant_sampling_parameters(**keyword_arguments)

    def test_multirange_fit_performance(self) -> None:
        """Verify that the fitting works with artificial dataset"""

        # First, generate a sum of modelled variograms: ranges and  partial sills for three models
        params_real = (100, 0.7, 1000, 0.2, 10000, 0.1)
        r1, ps1, r2, ps2, r3, ps3 = params_real

        x = np.linspace(10, 20000, 500)
        y = models.spherical(x, r=r1, c0=ps1) + models.spherical(x, r=r2, c0=ps2) + models.spherical(x, r=r3, c0=ps3)

        # Add some noise on top of it
        sig = 0.025
        np.random.seed(42)
        y_noise = np.random.normal(0, sig, size=len(x))

        y_simu = y + y_noise
        sigma = np.ones(len(x)) * sig

        # Put all in a dataframe
        df = pd.DataFrame()
        df = df.assign(lags=x, exp=y_simu, err_exp=sigma)

        # Run the fitting
        fun, params_est = xdem.spatialstats.fit_sum_model_variogram(["spherical", "spherical", "spherical"], df)

        for i in range(len(params_est)):
            # Assert all parameters were correctly estimated within a 30% relative margin
            assert params_real[2 * i] == pytest.approx(params_est["range"].values[i], rel=0.3)
            assert params_real[2 * i + 1] == pytest.approx(params_est["psill"].values[i], rel=0.3)

        if PLOT:
            xdem.spatialstats.plot_variogram(df, list_fit_fun=[fun])

    def test_check_params_variogram_model(self) -> None:
        """Verify that the checking function for the modelled variogram parameters dataframe returns adequate errors"""

        # Check when missing a column
        with pytest.raises(
            ValueError,
            match='The dataframe with variogram parameters must contain the columns "model",' ' "range" and "psill".',
        ):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["spherical"], "range": [100]})
            )

        # Check with wrong model format
        list_supported_models = ["spherical", "gaussian", "exponential", "cubic", "stable", "matern"]
        with pytest.raises(
            ValueError,
            match="Variogram model name Supraluminal not recognized. Supported models are: "
            + ", ".join(list_supported_models)
            + ".",
        ):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["Supraluminal"], "range": [100], "psill": [1]})
            )

        # Check with wrong range format
        with pytest.raises(ValueError, match="The variogram ranges must be float or integer."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["spherical"], "range": ["a"], "psill": [1]})
            )

        # Check with negative range
        with pytest.raises(ValueError, match="The variogram ranges must have non-zero, positive values."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["spherical"], "range": [-1], "psill": [1]})
            )

        # Check with wrong partial sill format
        with pytest.raises(ValueError, match="The variogram partial sills must be float or integer."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["spherical"], "range": [100], "psill": ["a"]})
            )

        # Check with negative partial sill
        with pytest.raises(ValueError, match="The variogram partial sills must have non-zero, positive values."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["spherical"], "range": [100], "psill": [-1]})
            )

        # Check with a model that requires smoothness and without the smoothness column
        with pytest.raises(
            ValueError,
            match='The dataframe with variogram parameters must contain the column "smooth" '
            "for the smoothness factor when using Matern or Stable models.",
        ):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["stable"], "range": [100], "psill": [1]})
            )

        # Check with wrong smoothness format
        with pytest.raises(ValueError, match="The variogram smoothness parameter must be float or integer."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["stable"], "range": [100], "psill": [1], "smooth": ["a"]})
            )

        # Check with negative smoothness
        with pytest.raises(ValueError, match="The variogram smoothness parameter must have non-zero, positive values."):
            xdem.spatialstats._check_validity_params_variogram(
                pd.DataFrame(data={"model": ["stable"], "range": [100], "psill": [1], "smooth": [-1]})
            )

    def test_estimate_model_spatial_correlation_and_infer_from_stable(self) -> None:
        """Test consistency of outputs and errors in wrapper functions for estimation of spatial correlation"""

        # Keep only data on stable
        diff_on_stable = self.diff.copy()
        diff_on_stable.set_mask(self.mask)

        # Load the error map from TestBinning
        errors = Raster(os.path.join(examples._EXAMPLES_DIRECTORY, "dh_error.tif"))

        # Standardize the differences
        zscores = diff_on_stable / errors

        # Run wrapper estimate and model function
        emp_vgm_1, params_model_vgm_1, _ = xdem.spatialstats.estimate_model_spatial_correlation(
            dvalues=zscores, list_models=["Gau", "Sph"], subsample=10, random_state=42
        )

        # Check that the output matches that of the original function under the same random state
        emp_vgm_2 = xdem.spatialstats.sample_empirical_variogram(
            values=zscores, estimator="dowd", subsample=10, random_state=42
        )
        pd.testing.assert_frame_equal(emp_vgm_1, emp_vgm_2)
        params_model_vgm_2 = xdem.spatialstats.fit_sum_model_variogram(
            list_models=["Gau", "Sph"], empirical_variogram=emp_vgm_2
        )[1]
        pd.testing.assert_frame_equal(params_model_vgm_1, params_model_vgm_2)

        # Run wrapper infer from stable function with a Raster and the mask, and check the consistency there as well
        emp_vgm_3, params_model_vgm_3, _ = xdem.spatialstats.infer_spatial_correlation_from_stable(
            dvalues=zscores, stable_mask=~self.mask, list_models=["Gau", "Sph"], subsample=10, random_state=42
        )
        pd.testing.assert_frame_equal(emp_vgm_1, emp_vgm_3)
        pd.testing.assert_frame_equal(params_model_vgm_1, params_model_vgm_3)

        # Run again with array instead of Raster as input
        zscores_arr = gu.raster.get_array_and_mask(zscores)[0]
        emp_vgm_4, params_model_vgm_4, _ = xdem.spatialstats.infer_spatial_correlation_from_stable(
            dvalues=zscores_arr,
            gsd=self.diff.res[0],
            stable_mask=~self.mask,
            list_models=["Gau", "Sph"],
            subsample=10,
            random_state=42,
        )
        pd.testing.assert_frame_equal(emp_vgm_1, emp_vgm_4)
        pd.testing.assert_frame_equal(params_model_vgm_1, params_model_vgm_4)

        # Run with a decent amount of samples to save to file
        _, params_model_vgm_5, _ = xdem.spatialstats.infer_spatial_correlation_from_stable(
            dvalues=zscores_arr,
            gsd=self.diff.res[0],
            stable_mask=~self.mask,
            list_models=["Gau", "Sph"],
            subsample=200,
            random_state=42,
        )
        # Save the modelled variogram for later used in TestNeffEstimation
        params_model_vgm_5.to_csv(
            os.path.join(examples._EXAMPLES_DIRECTORY, "df_variogram_model_params.csv"), index=False
        )

        # Check that errors are raised with wrong input
        with pytest.raises(ValueError, match="The values must be a Raster or NumPy array, or a list of those."):
            xdem.spatialstats.infer_spatial_correlation_from_stable(
                dvalues="not_an_array", stable_mask=~self.mask, list_models=["Gau", "Sph"], random_state=42
            )
        with pytest.raises(ValueError, match="The stable mask must be a Vector, Mask, GeoDataFrame or NumPy array."):
            xdem.spatialstats.infer_spatial_correlation_from_stable(
                dvalues=self.diff, stable_mask="not_a_vector_or_array", list_models=["Gau", "Sph"], random_state=42
            )
        with pytest.raises(ValueError, match="The unstable mask must be a Vector, Mask, GeoDataFrame or NumPy array."):
            xdem.spatialstats.infer_spatial_correlation_from_stable(
                dvalues=self.diff, unstable_mask="not_a_vector_or_array", list_models=["Gau", "Sph"], random_state=42
            )
        diff_on_stable_arr = gu.raster.get_array_and_mask(diff_on_stable)[0]
        with pytest.raises(
            ValueError,
            match="The stable mask can only passed as a Vector or GeoDataFrame if the input "
            "values contain a Raster.",
        ):
            xdem.spatialstats.infer_spatial_correlation_from_stable(
                dvalues=diff_on_stable_arr, stable_mask=self.outlines, list_models=["Gau", "Sph"], random_state=42
            )

    @pytest.mark.skip("Need to wait until SciPy publishes 1.11.3 with fix.")  # type: ignore
    def test_empirical_fit_plotting(self) -> None:
        """Verify that the shape of the empirical variogram output works with the fit and plotting"""

        # Check the variogram estimation runs for a random state
        df = xdem.spatialstats.sample_empirical_variogram(
            values=self.diff.data, gsd=self.diff.res[0], subsample=50, random_state=42
        )

        # Single model fit
        fun, _ = xdem.spatialstats.fit_sum_model_variogram(["spherical"], empirical_variogram=df)

        # Triple model fit
        fun2, _ = xdem.spatialstats.fit_sum_model_variogram(
            ["spherical", "spherical", "spherical"], empirical_variogram=df
        )

        if PLOT:
            # Plot with a single model fit
            xdem.spatialstats.plot_variogram(df, list_fit_fun=[fun])
            # Plot with a triple model fit
            xdem.spatialstats.plot_variogram(df, list_fit_fun=[fun2])

        # Check that errors are raised with wrong inputs
        # If the experimental variogram values "exp" are not passed
        with pytest.raises(
            ValueError, match='The expected variable "exp" is not part of the provided dataframe column names.'
        ):
            xdem.spatialstats.plot_variogram(pd.DataFrame(data={"wrong_name": [1], "lags": [1], "count": [100]}))
        # If the spatial lags "lags" are not passed
        with pytest.raises(
            ValueError, match='The expected variable "lags" is not part of the provided dataframe column names.'
        ):
            xdem.spatialstats.plot_variogram(pd.DataFrame(data={"exp": [1], "wrong_name": [1], "count": [100]}))
        # If the pairwise sample count "count" is not passed
        with pytest.raises(
            ValueError, match='The expected variable "count" is not part of the provided dataframe column names.'
        ):
            xdem.spatialstats.plot_variogram(pd.DataFrame(data={"exp": [1], "lags": [1], "wrong_name": [100]}))


class TestNeffEstimation:

    ref, diff, _, outlines = load_ref_and_diff()

    @pytest.mark.parametrize("range1", [10**i for i in range(3)])  # type: ignore
    @pytest.mark.parametrize("psill1", [0.1, 1, 10])  # type: ignore
    @pytest.mark.parametrize("model1", ["spherical", "exponential", "gaussian", "cubic"])  # type: ignore
    @pytest.mark.parametrize("area", [10 ** (2 * i) for i in range(3)])  # type: ignore
    def test_neff_circular_single_range(self, range1: float, psill1: float, model1: float, area: float) -> None:
        """Test the accuracy of numerical integration for one to three models of spherical, gaussian or exponential
        forms to get the number of effective samples"""

        params_variogram_model = pd.DataFrame(data={"model": [model1], "range": [range1], "psill": [psill1]})

        # Exact integration
        neff_circ_exact = xdem.spatialstats.neff_circular_approx_theoretical(
            area=area, params_variogram_model=params_variogram_model
        )
        # Numerical integration
        neff_circ_numer = xdem.spatialstats.neff_circular_approx_numerical(
            area=area, params_variogram_model=params_variogram_model
        )

        # Check results are the exact same
        assert neff_circ_exact == pytest.approx(neff_circ_numer, rel=0.001)

    @pytest.mark.parametrize("range1", [10**i for i in range(2)])  # type: ignore
    @pytest.mark.parametrize("range2", [10**i for i in range(2)])  # type: ignore
    @pytest.mark.parametrize("range3", [10**i for i in range(2)])  # type: ignore
    @pytest.mark.parametrize("model1", ["spherical", "exponential", "gaussian", "cubic"])  # type: ignore
    @pytest.mark.parametrize("model2", ["spherical", "exponential", "gaussian", "cubic"])  # type: ignore
    def test_neff_circular_three_ranges(
        self, range1: float, range2: float, range3: float, model1: float, model2: float
    ) -> None:
        """Test the accuracy of numerical integration for one to three models of spherical, gaussian or
        exponential forms"""

        area = 1000
        psill1 = 1
        psill2 = 1
        psill3 = 1
        model3 = "spherical"

        params_variogram_model = pd.DataFrame(
            data={
                "model": [model1, model2, model3],
                "range": [range1, range2, range3],
                "psill": [psill1, psill2, psill3],
            }
        )

        # Exact integration
        neff_circ_exact = xdem.spatialstats.neff_circular_approx_theoretical(
            area=area, params_variogram_model=params_variogram_model
        )
        # Numerical integration
        neff_circ_numer = xdem.spatialstats.neff_circular_approx_numerical(
            area=area, params_variogram_model=params_variogram_model
        )

        # Check results are the exact same
        assert neff_circ_exact == pytest.approx(neff_circ_numer, rel=0.001)

    def test_neff_exact_and_approx_hugonnet(self) -> None:
        """Test the exact and approximated calculation of the number of effective sample by double covariance sum"""

        # Generate a gridded dataset with varying errors associated to each pixel
        shape = (15, 15)
        errors = np.ones(shape)

        # Coordinates
        x = np.arange(0, shape[0])
        y = np.arange(0, shape[1])
        xx, yy = np.meshgrid(x, y)

        # Flatten everything
        coords = np.dstack((xx.ravel(), yy.ravel())).squeeze()
        errors = errors.ravel()

        # Create a list of variogram that, summed, represent the spatial correlation
        params_variogram_model = pd.DataFrame(
            data={"model": ["spherical", "gaussian"], "range": [5, 50], "psill": [0.5, 0.5]}
        )

        # Check that the function runs with default parameters
        # t0 = time.time()
        neff_exact = xdem.spatialstats.neff_exact(
            coords=coords, errors=errors, params_variogram_model=params_variogram_model
        )
        # t1 = time.time()

        # Check that the non-vectorized version gives the same result
        neff_exact_nv = xdem.spatialstats.neff_exact(
            coords=coords, errors=errors, params_variogram_model=params_variogram_model, vectorized=False
        )
        # t2 = time.time()
        assert neff_exact == pytest.approx(neff_exact_nv, rel=0.001)

        # Check that the vectorized version is faster (vectorized for about 250 points here)
        # assert (t1 - t0) < (t2 - t1)

        # Check that the approximation function runs with default parameters, sampling 100 out of 250 samples
        # t3 = time.time()
        neff_approx = xdem.spatialstats.neff_hugonnet_approx(
            coords=coords, errors=errors, params_variogram_model=params_variogram_model, subsample=100, random_state=42
        )
        # t4 = time.time()

        # Check that the non-vectorized version gives the same result, sampling 100 out of 250 samples
        neff_approx_nv = xdem.spatialstats.neff_hugonnet_approx(
            coords=coords,
            errors=errors,
            params_variogram_model=params_variogram_model,
            subsample=100,
            vectorized=False,
            random_state=42,
        )

        assert neff_approx == pytest.approx(neff_approx_nv, rel=0.001)

        # Check that the approximation version is faster within 30% error
        # TODO: find a more robust way to test time for CI
        # assert (t4 - t3) < (t1 - t0)

        # Check that the approximation is about the same as the original estimate within 10%
        assert neff_approx == pytest.approx(neff_exact, rel=0.1)

    def test_number_effective_samples(self) -> None:
        """Test that the wrapper function for neff functions behaves correctly and that output values are robust"""

        # The function should return the same result as neff_circular_approx_numerical when using a numerical area
        area = 10000
        params_variogram_model = pd.DataFrame(
            data={"model": ["spherical", "gaussian"], "range": [300, 3000], "psill": [0.5, 0.5]}
        )

        neff1 = xdem.spatialstats.neff_circular_approx_numerical(
            area=area, params_variogram_model=params_variogram_model
        )
        neff2 = xdem.spatialstats.number_effective_samples(area=area, params_variogram_model=params_variogram_model)

        assert neff1 == pytest.approx(neff2, rel=0.0001)

        # The function should return the same results as neff_hugonnet_approx when using a shape area
        # First, get the vector area and compute with the wrapper function
        res = 100.0
        outlines_brom = Vector(self.outlines.ds[self.outlines.ds["NAME"] == "Brombreen"])
        neff1 = xdem.spatialstats.number_effective_samples(
            area=outlines_brom,
            params_variogram_model=params_variogram_model,
            rasterize_resolution=res,
            random_state=42,
            subsample=10,
        )
        # Second, get coordinates manually and compute with the neff_approx_hugonnet function
        mask = outlines_brom.create_mask(xres=res, as_array=True)
        x = res * np.arange(0, mask.shape[0])
        y = res * np.arange(0, mask.shape[1])
        coords = np.array(np.meshgrid(y, x))
        coords_on_mask = coords[:, mask].T
        errors_on_mask = np.ones(len(coords_on_mask))
        neff2 = xdem.spatialstats.neff_hugonnet_approx(
            coords=coords_on_mask,
            errors=errors_on_mask,
            subsample=10,
            params_variogram_model=params_variogram_model,
            random_state=42,
        )
        # We can test the match between values accurately thanks to the random_state
        assert neff1 == pytest.approx(neff2, rel=0.00001)

        # Check that using a Raster as input for the resolution works
        neff3 = xdem.spatialstats.number_effective_samples(
            area=outlines_brom,
            subsample=10,
            params_variogram_model=params_variogram_model,
            rasterize_resolution=self.ref,
            random_state=42,
        )
        # The value should be nearly the same within 5% (the discretization grid is different so affects a tiny bit the
        # result)
        assert neff3 == pytest.approx(neff2, rel=0.05)

        # Check that the number of effective samples matches that of the circular approximation within 20%
        area_brom = np.sum(outlines_brom.ds.area.values)
        neff4 = xdem.spatialstats.number_effective_samples(
            area=area_brom, params_variogram_model=params_variogram_model
        )
        assert neff4 == pytest.approx(neff2, rel=0.2)
        # The circular approximation is always conservative, so should yield a smaller value
        assert neff4 < neff2

        # Check that errors are correctly raised
        with pytest.warns(
            UserWarning,
            match="Resolution for vector rasterization is not defined and thus set at 20% "
            "of the shortest correlation range, which might result in large memory usage.",
        ):
            xdem.spatialstats.number_effective_samples(
                area=outlines_brom, params_variogram_model=params_variogram_model
            )
        with pytest.raises(ValueError, match="Area must be a float, integer, Vector subclass or geopandas dataframe."):
            xdem.spatialstats.number_effective_samples(
                area="not supported", params_variogram_model=params_variogram_model
            )
        with pytest.raises(ValueError, match="The rasterize resolution must be a float, integer or Raster subclass."):
            xdem.spatialstats.number_effective_samples(
                area=outlines_brom, params_variogram_model=params_variogram_model, rasterize_resolution=(10, 10)
            )

    def test_spatial_error_propagation(self) -> None:
        """Test that the spatial error propagation wrapper function runs properly"""

        # Load the error map from TestBinning
        errors = Raster(os.path.join(examples._EXAMPLES_DIRECTORY, "dh_error.tif"))

        # Load the spatial correlation from TestVariogram
        params_variogram_model = pd.read_csv(
            os.path.join(examples._EXAMPLES_DIRECTORY, "df_variogram_model_params.csv"), index_col=None
        )

        # Run the function with vector areas
        areas_vector = [
            self.outlines.ds[self.outlines.ds["NAME"] == "Brombreen"],
            self.outlines.ds[self.outlines.ds["NAME"] == "Medalsbreen"],
        ]

        list_stderr_vec = xdem.spatialstats.spatial_error_propagation(
            areas=areas_vector,
            errors=errors,
            params_variogram_model=params_variogram_model,
            random_state=42,
            subsample=10,
        )

        # Run the function with numeric areas (sum needed for Medalsbreen that has two separate polygons)
        areas_numeric = [np.sum(area_vec.area.values) for area_vec in areas_vector]
        list_stderr = xdem.spatialstats.spatial_error_propagation(
            areas=areas_numeric, errors=errors, params_variogram_model=params_variogram_model
        )

        # Check that the outputs are consistent: the numeric method should always give a neff that is almost the same
        # (20% relative) for those two glaciers as their shape is not too different from a disk
        for i in range(2):
            assert list_stderr_vec[i] == pytest.approx(list_stderr[i], rel=0.2)


class TestSubSampling:
    def test_circular_masking(self) -> None:
        """Test that the circular masking works as intended"""

        # using default (center should be [2,2], radius 2)
        circ = xdem.spatialstats._create_circular_mask((5, 5))
        circ2 = xdem.spatialstats._create_circular_mask((5, 5), center=(2, 2), radius=2)

        # check default center and radius are derived properly
        assert np.array_equal(circ, circ2)

        # check mask
        # masking is not inclusive, i.e. exactly radius=2 won't include the 2nd pixel from the center, but radius>2 will
        eq_circ = np.zeros((5, 5), dtype=bool)
        eq_circ[1:4, 1:4] = True
        assert np.array_equal(circ, eq_circ)

        # check distance is not a multiple of pixels (more accurate subsampling)
        # will create a 1-pixel mask around the center
        circ3 = xdem.spatialstats._create_circular_mask((5, 5), center=(1, 1), radius=1)

        eq_circ3 = np.zeros((5, 5), dtype=bool)
        eq_circ3[1, 1] = True
        assert np.array_equal(circ3, eq_circ3)

        # will create a square mask (<1.5 pixel) around the center
        circ4 = xdem.spatialstats._create_circular_mask((5, 5), center=(1, 1), radius=1.5)
        # should not be the same as radius = 1
        assert not np.array_equal(circ3, circ4)

    def test_ring_masking(self) -> None:
        """Test that the ring masking works as intended"""
        warnings.simplefilter("error")

        # by default, the mask is only an outside circle (ring of size 0)
        ring1 = xdem.spatialstats._create_ring_mask((5, 5))
        circ1 = xdem.spatialstats._create_circular_mask((5, 5))

        assert np.array_equal(ring1, circ1)

        # test rings with different inner radius
        ring2 = xdem.spatialstats._create_ring_mask((5, 5), in_radius=1, out_radius=2)
        ring3 = xdem.spatialstats._create_ring_mask((5, 5), in_radius=0, out_radius=2)
        ring4 = xdem.spatialstats._create_ring_mask((5, 5), in_radius=1.5, out_radius=2)

        assert np.logical_and(~np.array_equal(ring2, ring3), ~np.array_equal(ring3, ring4))

        # check default
        eq_ring2 = np.zeros((5, 5), dtype=bool)
        eq_ring2[1:4, 1:4] = True
        eq_ring2[2, 2] = False
        assert np.array_equal(ring2, eq_ring2)


class TestPatchesMethod:
    def test_patches_method_loop_quadrant(self) -> None:
        """Check that the patches method with quadrant loops (vectorized=False) functions correctly"""

        diff, mask = load_ref_and_diff()[1:3]

        gsd = diff.res[0]
        area = 100000

        # Check the patches method runs
        df, df_full = xdem.spatialstats.patches_method(
            diff,
            unstable_mask=mask,
            gsd=gsd,
            areas=[area],
            random_state=42,
            n_patches=100,
            vectorized=False,
            return_in_patch_statistics=True,
        )

        # First, the summary dataframe
        assert df.shape == (1, 4)
        assert all(df.columns == ["nmad", "nb_indep_patches", "exact_areas", "areas"])

        # Check the sampling is fixed for a random state
        assert df["nmad"][0] == pytest.approx(1.8530521253631773, abs=1e-3)
        assert df["nb_indep_patches"][0] == 100
        assert df["exact_areas"][0] == pytest.approx(df["areas"][0], rel=0.2)

        # Then, the full dataframe
        assert df_full.shape == (100, 5)

        # Check the sampling is always fixed for a random state
        assert df_full["tile"].values[0] == "8_16"
        assert df_full["nanmean"].values[0] == pytest.approx(0.3085488928369729, abs=1e-3)

        # Check that all counts respect the default minimum percentage of 80% valid pixels
        assert all(df_full["count"].values > 0.8 * np.max(df_full["count"].values))

    def test_patches_method_convolution(self) -> None:
        """Check that the patches method with convolution (vectorized=True) functions correctly"""

        diff, mask = load_ref_and_diff()[1:3]

        gsd = diff.res[0]
        area = 100000

        # First, the patches method runs with scipy
        df = xdem.spatialstats.patches_method(
            diff,
            unstable_mask=mask,
            gsd=gsd,
            areas=[area, area * 10],
            random_state=42,
            vectorized=True,
            convolution_method="scipy",
        )

        assert df.shape == (2, 4)
        assert all(df.columns == ["nmad", "nb_indep_patches", "exact_areas", "areas"])
        assert df["exact_areas"][0] == pytest.approx(df["areas"][0], rel=0.2)

        # Second, with numba
        # df, df_full = xdem.spatialstats.patches_method(
        #     diff,
        #     unstable_mask=mask.squeeze(),
        #     gsd=gsd,
        #     areas=[area],
        #     random_state=42,
        #     vectorized=True,
        #     convolution_method='numba',
        #     return_in_patch_statistics=True)
        #
        # assert df.shape == (1, 4)
        # assert all(df.columns == ['nmad', 'nb_indep_patches', 'exact_areas', 'areas'])
        # assert df['exact_areas'][0] == pytest.approx(df['areas'][0], rel=0.2)



from __future__ import annotations

import os
import re
import tempfile
import warnings

import geoutils as gu
import numpy as np
import pytest

import xdem
from xdem._typing import MArrayf

xdem.examples.download_longyearbyen_examples()

PLOT = True


def run_gdaldem(filepath: str, processing: str, options: str | None = None) -> MArrayf:
    """Run GDAL's DEMProcessing and return the read numpy array."""
    # Rasterio strongly recommends against importing gdal along rio, so this is done here instead.
    from osgeo import gdal

    # Converting string into gdal processing options here to avoid import gdal outside this function:
    # Riley or Wilson for Terrain Ruggedness, and Zevenberg or Horn for slope, aspect and hillshade
    gdal_option_conversion = {
        "Riley": gdal.DEMProcessingOptions(alg="Riley"),
        "Wilson": gdal.DEMProcessingOptions(alg="Wilson"),
        "Zevenberg": gdal.DEMProcessingOptions(alg="ZevenbergenThorne"),
        "Horn": gdal.DEMProcessingOptions(alg="Horn"),
        "hillshade_Zevenberg": gdal.DEMProcessingOptions(azimuth=315, altitude=45, alg="ZevenbergenThorne"),
        "hillshade_Horn": gdal.DEMProcessingOptions(azimuth=315, altitude=45, alg="Horn"),
    }

    if options is None:
        gdal_option = gdal.DEMProcessingOptions(options=None)
    else:
        gdal_option = gdal_option_conversion[options]

    temp_dir = tempfile.TemporaryDirectory()
    temp_path = os.path.join(temp_dir.name, "output.tif")
    gdal.DEMProcessing(
        destName=temp_path,
        srcDS=filepath,
        processing=processing,
        options=gdal_option,
    )

    data = gu.Raster(temp_path).data
    temp_dir.cleanup()
    return data


class TestTerrainAttribute:
    filepath = xdem.examples.get_path("longyearbyen_ref_dem")

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Parse metadata")
        dem = xdem.DEM(filepath, silent=True)

    @pytest.mark.parametrize(
        "attribute",
        [
            "slope_Horn",
            "aspect_Horn",
            "hillshade_Horn",
            "slope_Zevenberg",
            "aspect_Zevenberg",
            "hillshade_Zevenberg",
            "tri_Riley",
            "tri_Wilson",
            "tpi",
            "roughness",
        ],
    )  # type: ignore
    def test_attribute_functions_against_gdaldem(self, attribute: str) -> None:
        """
        Test that all attribute functions give the same results as those of GDALDEM within a small tolerance.

        :param attribute: The attribute to test (e.g. 'slope')
        """
        # TODO: New warnings to remove with latest GDAL versions, opening issue
        # warnings.simplefilter("error")

        functions = {
            "slope_Horn": lambda dem: xdem.terrain.slope(dem.data, dem.res, degrees=True),
            "aspect_Horn": lambda dem: xdem.terrain.aspect(dem.data, degrees=True),
            "hillshade_Horn": lambda dem: xdem.terrain.hillshade(dem.data, dem.res),
            "slope_Zevenberg": lambda dem: xdem.terrain.slope(
                dem.data, dem.res, method="ZevenbergThorne", degrees=True
            ),
            "aspect_Zevenberg": lambda dem: xdem.terrain.aspect(dem.data, method="ZevenbergThorne", degrees=True),
            "hillshade_Zevenberg": lambda dem: xdem.terrain.hillshade(dem.data, dem.res, method="ZevenbergThorne"),
            "tri_Riley": lambda dem: xdem.terrain.terrain_ruggedness_index(dem.data, method="Riley"),
            "tri_Wilson": lambda dem: xdem.terrain.terrain_ruggedness_index(dem.data, method="Wilson"),
            "tpi": lambda dem: xdem.terrain.topographic_position_index(dem.data),
            "roughness": lambda dem: xdem.terrain.roughness(dem.data),
        }

        # Writing dictionary options here to avoid importing gdal outside the dedicated function
        gdal_processing_attr_option = {
            "slope_Horn": ("slope", "Horn"),
            "aspect_Horn": ("aspect", "Horn"),
            "hillshade_Horn": ("hillshade", "hillshade_Horn"),
            "slope_Zevenberg": ("slope", "Zevenberg"),
            "aspect_Zevenberg": ("aspect", "Zevenberg"),
            "hillshade_Zevenberg": ("hillshade", "hillshade_Zevenberg"),
            "tri_Riley": ("TRI", "Riley"),
            "tri_Wilson": ("TRI", "Wilson"),
            "tpi": ("TPI", None),
            "roughness": ("Roughness", None),
        }

        # Copy the DEM to ensure that the inter-test state is unchanged, and because the mask will be modified.
        dem = self.dem.copy()

        # Derive the attribute using both GDAL and xdem
        attr_xdem = functions[attribute](dem).squeeze()
        attr_gdal = run_gdaldem(
            self.filepath,
            processing=gdal_processing_attr_option[attribute][0],
            options=gdal_processing_attr_option[attribute][1],
        )

        # For hillshade, we round into an integer to match GDAL's output
        if attribute in ["hillshade_Horn", "hillshade_Zevenberg"]:
            attr_xdem = attr_xdem.astype("int").astype("float32")

        # We compute the difference and keep only valid values
        diff = (attr_xdem - attr_gdal).filled(np.nan)
        diff_valid = diff[np.isfinite(diff)]

        try:
            # Difference between xdem and GDAL attribute
            # Mean of attribute values to get an order of magnitude of the attribute unit
            magn = np.nanmean(np.abs(attr_xdem))

            # Check that the attributes are similar within a tolerance of a thousandth of the magnitude
            # For instance, slopes have an average magnitude of around 30 deg, so the tolerance is 0.030 deg
            if attribute in ["hillshade_Horn", "hillshade_Zevenberg"]:
                # For hillshade, check 0 or 1 difference due to integer rounding
                assert np.all(np.logical_or(diff_valid == 0.0, np.abs(diff_valid) == 1.0))

            elif attribute in ["aspect_Horn", "aspect_Zevenberg"]:
                # For aspect, check the tolerance within a 360 degree modulo due to the circularity of the variable
                diff_valid = np.mod(np.abs(diff_valid), 360)
                assert np.all(np.minimum(diff_valid, np.abs(360 - diff_valid)) < 10 ** (-3) * magn)
            else:
                # All attributes other than hillshade and aspect are non-circular floats, so we check within a tolerance
                assert np.all(np.abs(diff_valid < 10 ** (-3) * magn))

        except Exception as exception:

            if PLOT:
                import matplotlib.pyplot as plt

                # Plotting the xdem and GDAL attributes for comparison (plotting "diff" can also help debug)
                plt.subplot(121)
                plt.imshow(attr_gdal.squeeze())
                plt.colorbar()
                plt.subplot(122)
                plt.imshow(attr_xdem.squeeze())
                plt.colorbar()
                plt.show()

            raise exception

        # Introduce some nans
        dem.data.mask = np.zeros_like(dem.data, dtype=bool)
        dem.data.mask.ravel()[np.random.choice(dem.data.size, 50000, replace=False)] = True

        # Validate that this doesn't raise weird warnings after introducing nans.
        functions[attribute](dem)

    @pytest.mark.skip(
        "richdem wheels don't build on latest GDAL versions, " "need to circumvent that problem..."
    )  # type: ignore
    @pytest.mark.parametrize(
        "attribute",
        ["slope_Horn", "aspect_Horn", "hillshade_Horn", "curvature", "profile_curvature", "planform_curvature"],
    )  # type: ignore
    def test_attribute_functions_against_richdem(self, attribute: str) -> None:
        """
        Test that all attribute functions give the same results as those of RichDEM within a small tolerance.

        :param attribute: The attribute to test (e.g. 'slope')
        """
        warnings.simplefilter("error")

        # Functions for xdem-implemented methods
        functions_xdem = {
            "slope_Horn": lambda dem: xdem.terrain.slope(dem, dem.res, degrees=True),
            "aspect_Horn": lambda dem: xdem.terrain.aspect(dem.data, degrees=True),
            "hillshade_Horn": lambda dem: xdem.terrain.hillshade(dem.data, dem.res),
            "curvature": lambda dem: xdem.terrain.curvature(dem.data, dem.res),
            "profile_curvature": lambda dem: xdem.terrain.profile_curvature(dem.data, dem.res),
            "planform_curvature": lambda dem: xdem.terrain.planform_curvature(dem.data, dem.res),
        }

        # Functions for RichDEM wrapper methods
        functions_richdem = {
            "slope_Horn": lambda dem: xdem.terrain.slope(dem, degrees=True, use_richdem=True),
            "aspect_Horn": lambda dem: xdem.terrain.aspect(dem, degrees=True, use_richdem=True),
            "hillshade_Horn": lambda dem: xdem.terrain.hillshade(dem, use_richdem=True),
            "curvature": lambda dem: xdem.terrain.curvature(dem, use_richdem=True),
            "profile_curvature": lambda dem: xdem.terrain.profile_curvature(dem, use_richdem=True),
            "planform_curvature": lambda dem: xdem.terrain.planform_curvature(dem, use_richdem=True),
        }

        # Copy the DEM to ensure that the inter-test state is unchanged, and because the mask will be modified.
        dem = self.dem.copy()

        # Derive the attribute using both RichDEM and xdem
        attr_xdem = gu.raster.get_array_and_mask(functions_xdem[attribute](dem))[0].squeeze()
        attr_richdem = gu.raster.get_array_and_mask(functions_richdem[attribute](dem))[0].squeeze()

        # We compute the difference and keep only valid values
        diff = attr_xdem - attr_richdem
        diff_valid = diff[np.isfinite(diff)]

        try:
            # Difference between xdem and RichDEM attribute
            # Mean of attribute values to get an order of magnitude of the attribute unit
            magn = np.nanmean(np.abs(attr_xdem))

            # Check that the attributes are similar within a tolerance of a thousandth of the magnitude
            # For instance, slopes have an average magnitude of around 30 deg, so the tolerance is 0.030 deg
            if attribute in ["aspect_Horn"]:
                # For aspect, check the tolerance within a 360 degree modulo due to the circularity of the variable
                diff_valid = np.mod(np.abs(diff_valid), 360)
                assert np.all(np.minimum(diff_valid, np.abs(360 - diff_valid)) < 10 ** (-3) * magn)

            else:
                # All attributes other than aspect are non-circular floats, so we check within a tolerance
                # Here hillshade is not rounded as integer by our calculation, so no need to differentiate as with GDAL
                assert np.all(np.abs(diff_valid < 10 ** (-3) * magn))

        except Exception as exception:

            if PLOT:
                import matplotlib.pyplot as plt

                # Plotting the xdem and RichDEM attributes for comparison (plotting "diff" can also help debug)
                plt.subplot(221)
                plt.imshow(attr_richdem)
                plt.colorbar()
                plt.subplot(222)
                plt.imshow(attr_xdem)
                plt.colorbar()
                plt.subplot(223)
                plt.imshow(diff)
                plt.colorbar()
                plt.show()

            raise exception

        # Introduce some nans
        dem.data.mask = np.zeros_like(dem.data, dtype=bool)
        dem.data.mask.ravel()[np.random.choice(dem.data.size, 50000, replace=False)] = True

        # Validate that this doesn't raise weird warnings after introducing nans and that mask is preserved
        output = functions_richdem[attribute](dem)
        assert np.all(dem.data.mask == output.data.mask)

    def test_hillshade_errors(self) -> None:
        """Validate that the hillshade function raises appropriate errors."""
        # Try giving the hillshade invalid arguments.
        warnings.simplefilter("error")

        with pytest.raises(ValueError, match="Azimuth must be a value between 0 and 360"):
            xdem.terrain.hillshade(self.dem.data, self.dem.res, azimuth=361)

        with pytest.raises(ValueError, match="Altitude must be a value between 0 and 90"):
            xdem.terrain.hillshade(self.dem.data, self.dem.res, altitude=91)

        with pytest.raises(ValueError, match="z_factor must be a non-negative finite value"):
            xdem.terrain.hillshade(self.dem.data, self.dem.res, z_factor=np.inf)

    def test_hillshade(self) -> None:
        """Test hillshade-specific settings."""
        warnings.simplefilter("error")
        zfactor_1 = xdem.terrain.hillshade(self.dem.data, self.dem.res, z_factor=1.0)
        zfactor_10 = xdem.terrain.hillshade(self.dem.data, self.dem.res, z_factor=10.0)

        # A higher z-factor should be more variable than a low one.
        assert np.nanstd(zfactor_1) < np.nanstd(zfactor_10)

        low_altitude = xdem.terrain.hillshade(self.dem.data, self.dem.res, altitude=10)
        high_altitude = xdem.terrain.hillshade(self.dem.data, self.dem.res, altitude=80)

        # A low altitude should be darker than a high altitude.
        assert np.nanmean(low_altitude) < np.nanmean(high_altitude)

    @pytest.mark.parametrize(
        "name", ["curvature", "planform_curvature", "profile_curvature", "maximum_curvature"]
    )  # type: ignore
    def test_curvatures(self, name: str) -> None:
        """Test the curvature functions"""
        warnings.simplefilter("error")

        # Copy the DEM to ensure that the inter-test state is unchanged, and because the mask will be modified.
        dem = self.dem.copy()

        # Derive curvature without any gaps
        curvature = xdem.terrain.get_terrain_attribute(
            dem.data, attribute=name, resolution=dem.res, edge_method="nearest"
        )

        # Validate that the array has the same shape as the input and that all values are finite.
        assert curvature.shape == dem.data.shape
        try:
            assert np.all(np.isfinite(curvature))
        except Exception:
            import matplotlib.pyplot as plt

            plt.imshow(curvature.squeeze())
            plt.show()

        with pytest.raises(ValueError, match="Quadric surface fit requires the same X and Y resolution."):
            xdem.terrain.get_terrain_attribute(dem.data, attribute=name, resolution=(1.0, 2.0))

        # Introduce some nans
        dem.data.mask = np.zeros_like(dem.data, dtype=bool)
        dem.data.mask.ravel()[np.random.choice(dem.data.size, 50000, replace=False)] = True
        # Validate that this doesn't raise weird warnings after introducing nans.
        xdem.terrain.get_terrain_attribute(dem.data, attribute=name, resolution=dem.res)

    def test_get_terrain_attribute(self) -> None:
        """Test the get_terrain_attribute function by itself."""
        warnings.simplefilter("error")
        # Validate that giving only one terrain attribute only returns that, and not a list of len() == 1
        slope = xdem.terrain.get_terrain_attribute(self.dem.data, "slope", resolution=self.dem.res)
        assert isinstance(slope, np.ndarray)

        # Create three products at the same time
        slope2, _, hillshade = xdem.terrain.get_terrain_attribute(
            self.dem.data, ["slope", "aspect", "hillshade"], resolution=self.dem.res
        )

        # Create a hillshade using its own function
        hillshade2 = xdem.terrain.hillshade(self.dem.data, self.dem.res)

        # Validate that the "batch-created" hillshades and slopes are the same as the "single-created"
        assert np.array_equal(hillshade, hillshade2, equal_nan=True)
        assert np.array_equal(slope, slope2, equal_nan=True)

        # A slope map with a lower resolution (higher value) should have gentler slopes.
        slope_lowres = xdem.terrain.get_terrain_attribute(self.dem.data, "slope", resolution=self.dem.res[0] * 2)
        assert np.nanmean(slope) > np.nanmean(slope_lowres)

    @pytest.mark.skip(
        "richdem wheels don't build on latest GDAL versions, " "need to circumvent that problem..."
    )  # type: ignore
    def test_get_terrain_attribute_errors(self) -> None:
        """Test the get_terrain_attribute function raises appropriate errors."""

        # Below, re.escape() is needed to match expressions that have special characters (e.g., parenthesis, bracket)
        with pytest.raises(
            ValueError,
            match=re.escape("RichDEM can only compute the slope and aspect using the " "default method of Horn (1981)"),
        ):
            xdem.terrain.slope(self.dem, method="ZevenbergThorne", use_richdem=True)

        with pytest.raises(ValueError, match="To derive RichDEM attributes, the DEM passed must be a Raster object"):
            xdem.terrain.slope(self.dem.data, resolution=self.dem.res, use_richdem=True)

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Slope method 'DoesNotExist' is not supported. Must be one of: " "['Horn', 'ZevenbergThorne']"
            ),
        ):
            xdem.terrain.slope(self.dem.data, method="DoesNotExist")

        with pytest.raises(
            ValueError,
            match=re.escape("TRI method 'DoesNotExist' is not supported. Must be one of: " "['Riley', 'Wilson']"),
        ):
            xdem.terrain.terrain_ruggedness_index(self.dem.data, method="DoesNotExist")

    def test_raster_argument(self) -> None:

        slope, aspect = xdem.terrain.get_terrain_attribute(self.dem, attribute=["slope", "aspect"])

        assert slope != aspect

        assert type(slope) == type(aspect)
        assert all(isinstance(r, gu.Raster) for r in (aspect, slope, self.dem))

        assert slope.transform == self.dem.transform == aspect.transform
        assert slope.crs == self.dem.crs == aspect.crs

    def test_rugosity_jenness(self) -> None:
        """
        Test the rugosity with the same example as in Jenness (2004),
        https://doi.org/10.2193/0091-7648(2004)032[0829:CLSAFD]2.0.CO;2.
        """

        # Derive rugosity from the function
        dem = np.array([[190, 170, 155], [183, 165, 145], [175, 160, 122]], dtype="float32")

        # Derive rugosity
        rugosity = xdem.terrain.rugosity(dem, resolution=100.0)

        # Rugosity of Jenness (2004) example
        r = 10280.48 / 10000.0

        assert rugosity[1, 1] == pytest.approx(r, rel=10 ** (-4))

    # Loop for various elevation differences with the center
    @pytest.mark.parametrize("dh", np.linspace(0.01, 100, 10))  # type: ignore
    # Loop for different resolutions
    @pytest.mark.parametrize("resolution", np.linspace(0.01, 100, 10))  # type: ignore
    def test_rugosity_simple_cases(self, dh: float, resolution: float) -> None:
        """Test the rugosity calculation for simple cases."""
        warnings.simplefilter("error")

        # We here check the value for a fully symmetric case: the rugosity calculation can be simplified because all
        # eight triangles have the same surface area, see Jenness (2004).

        # Derive rugosity from the function
        dem = np.array([[1, 1, 1], [1, 1 + dh, 1], [1, 1, 1]], dtype="float32")

        rugosity = xdem.terrain.rugosity(dem, resolution=resolution)

        # Half surface length between the center and a corner cell (in 3D: accounting for elevation changes)
        side1 = np.sqrt(2 * resolution**2 + dh**2) / 2.0
        # Half surface length between the center and a side cell (in 3D: accounting for elevation changes)
        side2 = np.sqrt(resolution**2 + dh**2) / 2.0
        # Half surface length between the corner and side cell (no elevation changes on this side)
        side3 = resolution / 2.0

        # Formula for area A of one triangle
        s = (side1 + side2 + side3) / 2.0
        A = np.sqrt(s * (s - side1) * (s - side2) * (s - side3))

        # We sum the area of the eight triangles, and divide by the planimetric area (resolution squared)
        r = 8 * A / (resolution**2)

        # Check rugosity value is valid
        assert r == pytest.approx(rugosity[1, 1], rel=10 ** (-6))

    def test_get_quadric_coefficients(self) -> None:
        """Test the outputs and exceptions of the get_quadric_coefficients() function."""
        warnings.simplefilter("error")

        dem = np.array([[1, 1, 1], [1, 2, 1], [1, 1, 1]], dtype="float32")

        coefficients = xdem.terrain.get_quadric_coefficients(
            dem, resolution=1.0, edge_method="nearest", make_rugosity=True
        )

        # Check all coefficients are finite with an edge method
        assert np.all(np.isfinite(coefficients))

        # The 4th to last coefficient is the dem itself (could maybe be removed in the future as it is duplication..)
        assert np.array_equal(coefficients[-4, :, :], dem)

        # The middle pixel (index 1, 1) should be concave in the x-direction
        assert coefficients[3, 1, 1] < 0

        # The middle pixel (index 1, 1) should be concave in the y-direction
        assert coefficients[4, 1, 1] < 0

        with pytest.raises(ValueError, match="Invalid input array shape"):
            xdem.terrain.get_quadric_coefficients(dem.reshape((1, 1, -1)), 1.0)

        # Validate that when using the edge_method="none", only the one non-edge value is kept.
        coefs = xdem.terrain.get_quadric_coefficients(dem, resolution=1.0, edge_method="none")
        assert np.count_nonzero(np.isfinite(coefs[0, :, :])) == 1
        # When using edge wrapping, all coefficients should be finite.
        coefs = xdem.terrain.get_quadric_coefficients(dem, resolution=1.0, edge_method="wrap")
        assert np.count_nonzero(np.isfinite(coefs[0, :, :])) == 9



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# Allow conf.py to find the xdem module
sys.path.append(os.path.abspath("../.."))
sys.path.append(os.path.abspath("../../xdem/"))
sys.path.append(os.path.abspath(".."))

from sphinx_gallery.sorting import ExplicitOrder

import xdem

# -- Project information -----------------------------------------------------

project = "xDEM"
copyright = "2021, Erik Mannerfelt, Romain Hugonnet, Amaury Dehecq and others"
author = "Erik Mannerfelt, Romain Hugonnet, Amaury Dehecq and others"

# The full version, including alpha/beta/rc tags
release = xdem.__version__

os.environ["PYTHON"] = sys.executable


# -- General configuration ---------------------------------------------------

master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # Create the API documentation automatically
    "sphinx.ext.viewcode",  # Create the "[source]" button in the API to show the source code.
    "matplotlib.sphinxext.plot_directive",  # Render matplotlib figures from code.
    "sphinx.ext.autosummary",  # Create API doc summary texts from the docstrings.
    "sphinx.ext.inheritance_diagram",  # For class inheritance diagrams (see coregistration.rst).
    "sphinx.ext.graphviz",  # To render graphviz diagrams.
    "sphinx_design",  # To render nice blocks
    "sphinx_autodoc_typehints",  # Include type hints in the API documentation.
    "sphinxcontrib.programoutput",
    "sphinx_gallery.gen_gallery",  # Examples gallery
    "sphinx.ext.intersphinx",
    # "myst_parser",  !! Not needed with myst_nb !! # Form of Markdown that works with sphinx, used a lot by the Sphinx Book Theme
    "myst_nb",  # MySt for rendering Jupyter notebook in documentation
]

# For sphinx design to work properly
myst_enable_extensions = ["colon_fence", "dollarmath"]

# For myst-nb to find the Jupyter kernel (=environment) to run from
nb_kernel_rgx_aliases = {".*xdem.*": "python3"}
nb_execution_raise_on_error = True

# autosummary_generate = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "geoutils": ("https://geoutils.readthedocs.io/en/latest", None),
    "rasterio": ("https://rasterio.readthedocs.io/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "pyproj": ("https://pyproj4.github.io/pyproj/stable", None),
    "geopandas": ("https://geopandas.org/en/stable", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "rioxarray": ("https://corteva.github.io/rioxarray/stable/", None),
}

sphinx_gallery_conf = {
    "examples_dirs": [
        os.path.join(os.path.dirname(__file__), "../", "../", "examples/basic"),
        os.path.join(os.path.dirname(__file__), "../", "../", "examples/advanced"),
    ],  # path to your example scripts
    "gallery_dirs": ["basic_examples", "advanced_examples"],  # path to where to save gallery generated output
    "inspect_global_variables": True,  # Make links to the class/function definitions.
    "reference_url": {
        # The module you locally document uses None
        "xdem": None,
    },
    # directory where function/class granular galleries are stored
    "backreferences_dir": "gen_modules/backreferences",
    "doc_module": ("xdem",),  # which function/class levels are used to create galleries
    # 'subsection_order': ExplicitOrder([os.path.join(os.path.dirname(__file__), "../", "../", "examples", "basic"),
    #                                    os.path.join(os.path.dirname(__file__), "../", "../", "examples", "advanced")])
    "remove_config_comments": True,
    # To remove comments such as sphinx-gallery-thumbnail-number (only works in code, not in text)
}

extlinks = {
    "issue": ("https://github.com/GlacioHack/xdem/issues/%s", "GH"),
    "pull": ("https://github.com/GlacioHack/xdem/pull/%s", "PR"),
}

# For matplotlib figures generate with sphinx plot: (suffix, dpi)
plot_formats = [(".png", 400)]

# To avoid long path names in inheritance diagrams
inheritance_alias = {
    "geoutils.georaster.raster.Raster": "geoutils.Raster",
    "geoutils.georaster.raster.Mask": "geoutils.Mask",
    "geoutils.georaster.satimg.SatelliteImage": "geoutils.SatelliteImage",
    "geoutils.geovector.Vector": "geoutils.Vector",
    "xdem.dem.DEM": "xdem.DEM",
    "xdem.coreg.base.Coreg": "xdem.Coreg",
    "xdem.coreg.affine.AffineCoreg": "xdem.AffineCoreg",
    "xdem.coreg.biascorr.BiasCorr": "xdem.BiasCorr",
}

# To have an edge color that works in both dark and light mode
inheritance_edge_attrs = {"color": "dodgerblue1"}

# To avoid fuzzy PNGs
graphviz_output_format = "svg"

# Add any paths that contain templates here, relative to this directory.
templates_path = [os.path.join(os.path.dirname(__file__), "_templates")]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_templates"]


# To ignore warnings due to having myst-nb reading the .ipynb created by sphinx-gallery
# Should eventually be fixed, see: https://github.com/executablebooks/MyST-NB/issues/363
def setup(app):
    # Ignore .ipynb files
    app.registry.source_suffix.pop(".ipynb", None)


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_favicon = "_static/xdem_logo_only.svg"
html_logo = "_static/xdem_logo.svg"
html_title = "xDEM"

html_theme_options = {
    "path_to_docs": "doc/source",
    "use_sidenotes": True,
    "repository_url": "https://github.com/GlacioHack/xdem",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_edit_page_button": True,
    "use_source_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org/",
        "notebook_interface": "jupyterlab",
        # For launching Binder in Jupyterlab to open MD files as notebook (downloads them otherwise)
    },
    "show_toc_level": 2,  # To show more levels on the right sidebar TOC
    "logo": {
        "image_dark": "_static/xdem_logo_dark.svg",
    },
}

# For dark mode
html_context = {
    # ...
    "default_mode": "auto"
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["imgs", "_static"]  # Commented out as we have no custom static data
